##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors: Kyle A Beauchamp
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


import numpy as np

from mdtraj.geometry import _geometry, distance
from mdtraj.utils import ensure_type

__all__ = ["compute_angles"]


def compute_angles(traj, angle_indices, periodic=True, opt=True):
    """Compute the bond angles between the supplied triplets of indices in each frame of a trajectory.

    Parameters
    ----------
    traj : Trajectory
        An mdtraj trajectory.
    angle_indices : np.ndarray, shape=(num_angles, 3), dtype=int
       Each row gives the indices of three atoms which together make an angle.
    periodic : bool, default=True
        If `periodic` is True and the trajectory contains unitcell
        information, we will treat angles that cross periodic images using
        the minimum image convention.
    opt : bool, default=True
        Use an optimized native library to calculate distances. Our optimized
        SSE angle calculation implementation is 10-20x faster than the
        (itself optimized) numpy implementation.

    Returns
    -------
    angles : np.ndarray, shape=[n_frames, n_angles], dtype=float
        The angles are in radians
    """
    xyz = ensure_type(
        traj.xyz,
        dtype=np.float32,
        ndim=3,
        name="traj.xyz",
        shape=(None, None, 3),
        warn_on_cast=False,
    )
    triplets = ensure_type(
        angle_indices,
        dtype=np.int32,
        ndim=2,
        name="angle_indices",
        shape=(None, 3),
        warn_on_cast=False,
    )
    if not np.all(np.logical_and(triplets < traj.n_atoms, triplets >= 0)):
        raise ValueError("angle_indices must be between 0 and %d" % traj.n_atoms)

    if len(triplets) == 0:
        return np.zeros((len(xyz), 0), dtype=np.float32)

    out = np.zeros((xyz.shape[0], triplets.shape[0]), dtype=np.float32)
    if periodic is True and traj._have_unitcell:
        box = ensure_type(
            traj.unitcell_vectors,
            dtype=np.float32,
            ndim=3,
            name="unitcell_vectors",
            shape=(len(xyz), 3, 3),
            warn_on_cast=False,
        )

        if opt:
            orthogonal = np.allclose(traj.unitcell_angles, 90)
            _geometry._angle_mic(
                xyz,
                triplets,
                box.transpose(0, 2, 1).copy(),
                out,
                orthogonal,
            )
            return out
        else:
            _angle(traj, triplets, periodic, out)
            return out

    if opt:
        _geometry._angle(xyz, triplets, out)
    else:
        _angle(traj, triplets, periodic, out)
    return out


def _angle(traj, angle_indices, periodic, out):
    ix01 = angle_indices[:, [1, 0]]
    ix21 = angle_indices[:, [1, 2]]

    u_prime = distance.compute_displacements(traj, ix01, periodic=periodic, opt=False)
    v_prime = distance.compute_displacements(traj, ix21, periodic=periodic, opt=False)
    u_norm = np.sqrt((u_prime**2).sum(-1))
    v_norm = np.sqrt((v_prime**2).sum(-1))

    # adding a new axis makes sure that broasting rules kick in on the third
    # dimension
    u = u_prime / (u_norm[..., np.newaxis])
    v = v_prime / (v_norm[..., np.newaxis])
    # use np.clip() to guard against errors when angles approach 180 degrees:
    return np.arccos(np.clip((u * v).sum(-1), -1.0, 1.0), out=out)



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2017 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors: Jason Swails, Matthew Harrigan
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


import os
import re
import tempfile

import numpy as np
import pytest
from conftest import flaky_pdb_dl

from mdtraj import Topology, load, load_pdb
from mdtraj.core.topology import float_to_bond_type
from mdtraj.formats.pdb import pdbstructure
from mdtraj.formats.pdb.pdbstructure import PdbStructure
from mdtraj.testing import eq

fd, temp = tempfile.mkstemp(suffix=".pdb")
os.close(fd)


def teardown_module(module):
    """remove the temporary file created by tests in this file
    this gets automatically called by pytest"""
    os.unlink(temp)


def test_pdbread(get_fn):
    load(get_fn("native.pdb"))


def test_pdbread_with_input_top(get_fn):
    pdb = get_fn("native.pdb")
    p_1 = load(pdb)

    p_2 = load(pdb, top=pdb)

    eq(p_1.xyz, p_2.xyz)
    eq(p_1.topology, p_2.topology)


def test_pdbwrite(get_fn):
    pdb = get_fn("native.pdb")
    p = load(pdb)
    p.save(temp)

    r = load(temp)
    eq(p.xyz, r.xyz)


def test_load_multiframe(get_fn):
    with open(get_fn("multiframe.pdb")) as f:
        pdb = PdbStructure(f)
        assert eq(len(pdb.models), 2)
        assert eq(len(pdb.models[0].chains), 1)
        assert eq(len(pdb.models[0].chains[0].residues), 3)
        assert eq(len(list(pdb.models[0].iter_atoms())), 22)

        assert eq(len(pdb.models[1].chains), 1)
        assert eq(len(pdb.models[1].chains[0].residues), 3)
        assert eq(len(list(pdb.models[1].iter_atoms())), 22)

    t = load(get_fn("multiframe.pdb"))
    assert eq(t.n_frames, 2)
    assert eq(t.n_atoms, 22)
    assert eq(t.xyz[0], t.xyz[1])


def test_4ZUO(get_fn):
    t = load(get_fn("4ZUO.pdb"))
    eq(t.n_frames, 1)
    eq(t.n_atoms, 6200)

    # this is a random line from the file
    # ATOM   1609  O   GLY A 201     -25.423  13.774 -25.234  1.00  8.92           O

    atom = list(t.top.atoms)[1525]
    eq(atom.element.symbol, "O")
    eq(atom.residue.name, "GLY")
    eq(atom.residue.chain.chain_id, "A")
    eq(atom.index, 1525)
    eq(t.xyz[0, 1525], np.array([-25.423, 13.774, -25.234]) / 10)  # converting to NM

    # this is atom 1577 in the PDB
    # CONECT 1577 5518
    # ATOM   1577  O   HIS A 197     -18.521   9.724 -32.805  1.00  8.81           O
    # HETATM 5518  K     K A 402     -19.609  10.871 -35.067  1.00  9.11           K
    atom = list(t.top.atoms)[1493]
    eq(atom.name, "O")
    eq(atom.residue.name, "HIS")
    eq(
        [(a1.index, a2.index) for a1, a2 in t.top.bonds if a1.index == 1493 or a2.index == 1493],
        [(1492, 1493), (1493, 5129)],
    )

    # that first bond is from a conect record


def test_2EQQ_0(get_fn):
    # this is an nmr structure with 20 models
    t = load(get_fn("2EQQ.pdb"))
    assert eq(t.n_frames, 20)
    assert eq(t.n_atoms, 423)
    assert eq(len(list(t.top.residues)), 28)


def test_1vii_solvated_with_ligand(get_fn):
    traj = load(get_fn("1vii_sustiva_water.pdb"))
    eq(len(list(traj.top.bonds)), 5156)
    eq(len([bond for bond in traj.top.bonds if bond[0].residue.name == "LIG"]), 32)
    traj.save(temp)
    traj = load(temp)
    eq(len(list(traj.top.bonds)), 5156)
    eq(len([bond for bond in traj.top.bonds if bond[0].residue.name == "LIG"]), 32)


def test_write_large(get_fn):
    traj = load(get_fn("native.pdb"))
    traj.xyz.fill(123456789)
    with pytest.raises(ValueError):
        traj.save(temp)


def test_write_large_2(get_fn):
    traj = load(get_fn("native.pdb"))
    traj.xyz.fill(-123456789)
    with pytest.raises(ValueError):
        traj.save(temp)


def test_pdbstructure_0():
    pdb_lines = [
        "ATOM    188  N   CYS A  42      40.714  -5.292  12.123  1.00 11.29           N  ",
        "ATOM    189  CA  CYS A  42      39.736  -5.883  12.911  1.00 10.01           C  ",
        "ATOM    190  C   CYS A  42      40.339  -6.654  14.087  1.00 22.28           C  ",
        "ATOM    191  O   CYS A  42      41.181  -7.530  13.859  1.00 13.70           O  ",
        "ATOM    192  CB  CYS A  42      38.949  -6.825  12.002  1.00  9.67           C  ",
        "ATOM    193  SG  CYS A  42      37.557  -7.514  12.922  1.00 20.12           S  ",
    ]

    res = pdbstructure.Residue("CYS", 42)
    for line in pdb_lines:
        res._add_atom(pdbstructure.Atom(line))
    for i, atom in enumerate(res):
        eq(pdb_lines[i], str(atom))


def test_pdbstructure_1():
    pdb_lines = [
        "ATOM    188  N   CYS A  42      40.714  -5.292  12.123  1.00 11.29           N",
        "ATOM    189  CA  CYS A  42      39.736  -5.883  12.911  1.00 10.01           C",
        "ATOM    190  C   CYS A  42      40.339  -6.654  14.087  1.00 22.28           C",
        "ATOM    191  O   CYS A  42      41.181  -7.530  13.859  1.00 13.70           O",
        "ATOM    192  CB  CYS A  42      38.949  -6.825  12.002  1.00  9.67           C",
        "ATOM    193  SG  CYS A  42      37.557  -7.514  12.922  1.00 20.12           S",
    ]
    positions = np.array(
        [
            [40.714, -5.292, 12.123],
            [39.736, -5.883, 12.911],
            [40.339, -6.654, 14.087],
            [41.181, -7.53, 13.859],
            [38.949, -6.825, 12.002],
            [37.557, -7.514, 12.922],
        ],
    )

    res = pdbstructure.Residue("CYS", 42)
    for line in pdb_lines:
        res._add_atom(pdbstructure.Atom(line))

    for i, c in enumerate(res.iter_positions()):
        eq(c, positions[i])


def test_pdbstructure_2():
    atom = pdbstructure.Atom(
        "ATOM   2209  CB  TYR A 299       6.167  22.607  20.046  1.00  8.12           C",
    )
    expected = np.array([6.167, 22.607, 20.046])
    for i, c in enumerate(atom.iter_coordinates()):
        eq(expected[i], c)


def test_pdbstructure_3():
    loc = pdbstructure.Atom.Location(" ", [1, 2, 3], 1.0, 20.0, "XXX")
    expected = [1, 2, 3]
    for i, c in enumerate(loc):
        eq(expected[i], c)


@flaky_pdb_dl
def test_load_pdb_from_url():
    # load pdb from URL
    t1 = load_pdb("https://www.rcsb.org/pdb/files/4ZUO.pdb.gz")
    t2 = load_pdb("https://www.rcsb.org/pdb/files/4ZUO.pdb")
    eq(t1.n_frames, 1)
    eq(t2.n_frames, 1)
    eq(t1.n_atoms, 6200)
    eq(t2.n_atoms, 6200)


@flaky_pdb_dl
def test_load_from_url():
    # load pdb from URL
    t1 = load("https://www.rcsb.org/pdb/files/4ZUO.pdb.gz")
    t2 = load("https://www.rcsb.org/pdb/files/4ZUO.pdb")
    eq(t1.n_frames, 1)
    eq(t2.n_frames, 1)
    eq(t1.n_atoms, 6200)
    eq(t2.n_atoms, 6200)


def test_3nch_conect(get_fn):
    # This has conect entries that use all available digits, good failure case.
    t1 = load_pdb(get_fn("3nch.pdb.gz"))
    top, bonds = t1.top.to_dataframe()
    bonds = {(a, b): 1 for (a, b, _, _) in bonds}
    eq(bonds[19782, 19783], 1)  # Check that last SO4 molecule has right bonds
    eq(bonds[19782, 19784], 1)  # Check that last SO4 molecule has right bonds
    eq(bonds[19782, 19785], 1)  # Check that last SO4 molecule has right bonds
    eq(bonds[19782, 19786], 1)  # Check that last SO4 molecule has right bonds


def test_3nch_serial_resSeq(get_fn):
    # If you use zero-based indexing, this PDB has quite large gaps in residue and atom numbering,
    # so it's a good test case.  See #528
    # Gold standard values obtained via
    # cat 3nch.pdb |grep ATM|tail -n 5
    # HETATM19787  S   SO4 D 804      -4.788  -9.395  22.515  1.00121.87           S
    # HETATM19788  O1  SO4 D 804      -3.815  -9.511  21.425  1.00105.97           O
    # HETATM19789  O2  SO4 D 804      -5.989  -8.733  21.999  1.00116.13           O
    # HETATM19790  O3  SO4 D 804      -5.130 -10.726  23.043  1.00108.74           O
    # HETATM19791  O4  SO4 D 804      -4.210  -8.560  23.575  1.00112.54           O
    t1 = load_pdb(get_fn("3nch.pdb.gz"))
    top, bonds = t1.top.to_dataframe()

    top2 = Topology.from_dataframe(top, bonds)
    eq(t1.top, top2)

    top = top.set_index("serial")  # Index by the actual data in the PDB
    eq(str(top.loc[19791]["name"]), "O4")
    eq(str(top.loc[19787]["name"]), "S")
    eq(str(top.loc[19787]["resName"]), "SO4")
    eq(int(top.loc[19787]["resSeq"]), 804)


def test_1ncw(get_fn):
    load_pdb(get_fn("1ncw.pdb.gz"))


@flaky_pdb_dl
def test_1vii_url_and_gz(get_fn):
    t1 = load_pdb("https://www.rcsb.org/pdb/files/1vii.pdb.gz")
    t2 = load_pdb("https://www.rcsb.org/pdb/files/1vii.pdb")
    t3 = load_pdb(get_fn("1vii.pdb.gz"))
    t4 = load_pdb(get_fn("1vii.pdb"))
    eq(t1.n_frames, 1)
    eq(t1.n_frames, t2.n_frames)
    eq(t1.n_frames, t3.n_frames)
    eq(t1.n_frames, t4.n_frames)

    eq(t1.n_atoms, t2.n_atoms)
    eq(t1.n_atoms, t3.n_atoms)
    eq(t1.n_atoms, t4.n_atoms)


@flaky_pdb_dl
def test_1vii_load_from_mixture(get_fn):
    # load pdb from URL and locally
    t1 = load(["https://www.rcsb.org/pdb/files/1vii.pdb.gz", get_fn("1vii.pdb.gz")])
    t2 = load_pdb(get_fn("1vii.pdb"))
    t3 = load([get_fn("1vii.pdb"), "https://www.rcsb.org/pdb/files/1vii.pdb", get_fn("1vii.pdb")])

    eq(t1.n_frames, 2)
    eq(t2.n_frames, 1)
    eq(t3.n_frames, 3)

    eq(t1.n_atoms, t2.n_atoms)
    eq(t1.n_atoms, t3.n_atoms)


def test_segment_id(get_fn):
    pdb = load_pdb(get_fn("ala_ala_ala.pdb"))
    pdb.save_pdb(temp)
    pdb2 = load_pdb(temp)

    correct_segment_id = "AAL"
    # check that all segment ids are set correctly
    for ridx, r in enumerate(pdb.top.residues):
        assert r.segment_id == correct_segment_id, (
            "residue %i (0-indexed) does not have segment_id set correctly from ala_ala_ala.pdb" % (ridx)
        )

    # check that all segment ids are set correctly after a new pdb file is written
    for ridx, (r1, r2) in enumerate(zip(pdb.top.residues, pdb2.top.residues)):
        assert r1.segment_id == r2.segment_id, (
            f"segment_id of residue {ridx} (0-indexed) in ala_ala_ala.pdb does not "
            "agree with value in after being written out to a new pdb file"
        )


def test_bfactors(get_fn):
    pdb = load_pdb(get_fn("native.pdb"))
    bfactors0 = np.arange(pdb.n_atoms) / 2.0 - 4.0  # (Get some decimals..)

    pdb.save_pdb(temp, bfactors=bfactors0)

    with open(temp) as fh:
        atom_lines = [line for line in fh.readlines() if re.search(r"^ATOM", line)]

    str_bfactors1 = [line[60:66] for line in atom_lines]
    flt_bfactors1 = np.array([float(i) for i in str_bfactors1])

    # check formatting has a space at the beginning and not at the end
    frmt = np.array([(s[0] == " ") and (s[-1] != " ") for s in str_bfactors1])
    assert np.all(frmt)

    # make sure the numbers are actually the same
    eq(bfactors0, flt_bfactors1)


def test_hex(get_fn):
    pdb = load_pdb(get_fn("water_hex.pdb.gz"))
    assert pdb.n_atoms == 100569
    assert pdb.n_residues == 33523
    pdb.save(temp)


def test_dummy_pdb_box_detection(get_fn, recwarn):
    traj = load(get_fn("2koc.pdb"))
    assert len(recwarn) == 1
    w = recwarn.pop(UserWarning)
    assert "Unlikely unit cell" in str(w.message)
    assert traj.unitcell_lengths is None, "Expected dummy box to be deleted"


def test_multichain_load_cycle(get_fn):
    # Issue 1611, make sure that save/load works for more than 1 chain
    pdb = load(get_fn("issue_1611.pdb"))
    bonds = [(bond.atom1.index, bond.atom2.index) for bond in pdb.topology.bonds]
    pdb.save(temp)
    pdb2 = load_pdb(temp)
    bonds2 = [(bond.atom1.index, bond.atom2.index) for bond in pdb2.topology.bonds]
    assert len(bonds) == len(bonds2)


def test_multichain_load_cycle_noter(get_fn):
    # Issue 1611, make sure that save/load works for more than 1 chain
    pdb = load(get_fn("issue_1611.pdb"))
    bonds = [(bond.atom1.index, bond.atom2.index) for bond in pdb.topology.bonds]
    pdb.save(temp, ter=False)
    pdb2 = load_pdb(temp)
    bonds2 = [(bond.atom1.index, bond.atom2.index) for bond in pdb2.topology.bonds]
    assert len(bonds) == len(bonds2)


def test_load_pdb_input_top(get_fn):
    pdb = get_fn("native.pdb")
    p_1 = load_pdb(pdb)

    p_2 = load_pdb(pdb, top=p_1.topology)

    eq(p_1.xyz, p_2.xyz)
    eq(p_1.topology, p_2.topology)


def test_chimera_indexing(get_fn):
    traj = load_pdb(get_fn("chimera_indexing.pdb"))  # this should just not fail

    assert traj.n_atoms == 9
    assert traj.n_residues == 3
    assert traj.topology._atoms[3].serial == 100000
    assert traj.topology._atoms[3].residue.resSeq == 10000
    assert traj.topology.n_bonds == 6
    assert traj.topology._atoms[6].serial == 100003
    assert traj.topology._atoms[6].residue.resSeq == 10001

    test_pos = np.array(
        [
            [10.613, 0.225, 12.764],
            [10.629, 0.313, 12.729],
            [10.596, 0.172, 12.686],
            [10.613, 0.225, 12.764],
            [10.629, 0.313, 12.729],
            [10.596, 0.172, 12.686],
            [10.613, 0.225, 12.764],
            [10.629, 0.313, 12.729],
            [10.596, 0.172, 12.686],
        ],
        dtype=np.float32,
    )
    assert np.array_equal(traj._xyz[0], test_pos)


def test_chimera_indexing_skip(get_fn):
    traj = load_pdb(get_fn("chimera_indexing_skip.pdb"))  # this should just not fail

    assert traj.n_atoms == 6
    assert traj.n_residues == 2
    assert traj.topology._atoms[3].serial == 100003
    assert traj.topology._atoms[3].residue.resSeq == 10001
    assert traj.topology.n_bonds == 4

    test_pos = np.array(
        [
            [10.613, 0.225, 12.764],
            [10.629, 0.313, 12.729],
            [10.596, 0.172, 12.686],
            [10.613, 0.225, 12.764],
            [10.629, 0.313, 12.729],
            [10.596, 0.172, 12.686],
        ],
        dtype=np.float32,
    )
    assert np.array_equal(traj._xyz[0], test_pos)


def test_vmd_indexing(get_fn):
    traj = load_pdb(get_fn("vmd_indexing.pdb"))  # this should just not fail

    assert traj.n_atoms == 8
    assert traj.n_residues == 6
    assert traj.topology._atoms[1].residue.resSeq == 2710
    assert traj.topology._atoms[3].serial == 100000
    assert traj.topology._atoms[3].residue.resSeq == 10000
    assert traj.topology.n_bonds == 2
    assert traj.topology._atoms[6].serial == 100003
    assert traj.topology._atoms[6].residue.resSeq == 10001
    assert traj.topology._atoms[7].serial == 100465
    assert traj.topology._atoms[7].residue.resSeq == 11000

    test_pos = np.array(
        [
            [10.613, 0.225, 12.764],
            [10.613, 0.225, 12.764],
            [10.613, 0.225, 12.764],
            [10.613, 0.225, 12.764],
            [10.629, 0.313, 12.729],
            [10.596, 0.172, 12.686],
            [10.613, 0.225, 12.764],
            [10.629, 0.313, 12.729],
        ],
        dtype=np.float32,
    )
    assert np.array_equal(traj._xyz[0], test_pos)


def test_overflow_indexing(get_fn):
    traj = load_pdb(get_fn("overflow_indexing.pdb"))  # this should just not fail

    assert traj.n_atoms == 9
    assert traj.n_residues == 3
    assert traj.topology._atoms[3].serial == 100000
    assert traj.topology._atoms[3].residue.resSeq == 10000
    assert traj.topology.n_bonds == 6
    assert traj.topology._atoms[6].serial == 100003
    assert traj.topology._atoms[6].residue.resSeq == 10001

    test_pos = np.array(
        [
            [10.613, 0.225, 12.764],
            [10.629, 0.313, 12.729],
            [10.596, 0.172, 12.686],
            [10.613, 0.225, 12.764],
            [10.629, 0.313, 12.729],
            [10.596, 0.172, 12.686],
            [10.613, 0.225, 12.764],
            [10.629, 0.313, 12.729],
            [10.596, 0.172, 12.686],
        ],
        dtype=np.float32,
    )
    assert np.array_equal(traj._xyz[0], test_pos)


def test_blank_indexing(get_fn):
    with pytest.warns(UserWarning):
        traj = load_pdb(get_fn("blank_indexing.pdb"))  # this should just not fail

    assert traj.n_atoms == 8
    assert traj.n_residues == 6
    assert traj.topology._atoms[1].residue.resSeq == 2710
    assert traj.topology._atoms[3].serial == 100000
    assert traj.topology._atoms[3].residue.resSeq == 10000
    assert traj.topology.n_bonds == 2
    assert traj.topology._atoms[6].serial == 100003
    assert traj.topology._atoms[6].residue.resSeq == 10001
    assert traj.topology._atoms[7].serial == 100004
    assert traj.topology._atoms[7].residue.resSeq == 10002

    test_pos = np.array(
        [
            [10.613, 0.225, 12.764],
            [10.613, 0.225, 12.764],
            [10.613, 0.225, 12.764],
            [10.613, 0.225, 12.764],
            [10.629, 0.313, 12.729],
            [10.596, 0.172, 12.686],
            [10.613, 0.225, 12.764],
            [10.629, 0.313, 12.729],
        ],
        dtype=np.float32,
    )
    assert np.array_equal(traj._xyz[0], test_pos)


def test_pdb_charge_read(get_fn):
    """Test that formal charges are read from PDB file"""

    traj = load_pdb(get_fn("1ply_charge.pdb"))

    charges = [atom.formal_charge for atom in traj.topology.atoms if atom.formal_charge is not None]

    # we expect 10 a list of length 10
    assert len(charges) == 10

    # all of the charges should be 1
    assert set(charges) == {1.0}


def test_pdb_charge_write(get_fn):
    """Test that formal charges are written to PDB file"""

    traj = load_pdb(get_fn("1ply_charge.pdb"))

    # write the trajectory to a new file
    traj.save_pdb(temp)

    # read the new file
    traj2 = load_pdb(temp)

    charges = [atom.formal_charge for atom in traj2.topology.atoms if atom.formal_charge is not None]

    # we expect 10 a list of length 10
    assert len(charges) == 10

    # all of the charges should be 1
    assert set(charges) == {1.0}


@pytest.mark.parametrize(
    "bond_orders, ref_orders",
    [
        (False, [None] * 32),
        (True, [None, 2] + [None] * 11 + [2] + [None] * 17 + [1]),
    ],
    ids=[
        "bond_orders=False",
        "bond_orders=True",
    ],
)
def test_ala3_bond_order_read(get_fn, bond_orders, ref_orders):
    """Test that bond orders/types are read properly from CONECT section of PDB file"""

    traj = load_pdb(get_fn("ala_ala_ala.pdb"), bond_orders=bond_orders)

    # Checking bond orders
    bo = [bond.order for bond in traj.top._bonds]
    eq(bo, ref_orders, err_msg="bond orders do not match reference")

    # Checking bond types
    bt = [bond.type for bond in traj.top._bonds]
    bt_from_bo = [float_to_bond_type(order) for order in bo]
    ref_bt = [float_to_bond_type(order) for order in ref_orders]

    # Make sure bond order and bond type match
    eq(bt_from_bo, bt, err_msg="bond types and bond orders don't match")
    # Make sure bond type match reference
    eq(bt, ref_bt, err_msg="bond types do not match reference")


@pytest.mark.parametrize(
    "bond_orders, ref_orders",
    [
        (False, [1] * 12),
        (True, [2, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 1]),
    ],
    ids=["omit_write_bond_order", "write_bond_order"],
)
def test_bnz_bond_order_conect_write(get_fn, bond_orders, ref_orders):
    """Test that formal charges are written to PDB file"""

    traj = load_pdb(get_fn("bnz.pdb"), bond_orders=True)

    # write the trajectory to a new file
    traj.save_pdb(temp, bond_orders=bond_orders)

    # read the new file, always with bond order info
    traj2 = load_pdb(temp, bond_orders=True)

    bo = [bond.order for bond in traj2.top._bonds]
    eq(bo, ref_orders, err_msg="Inconsistent number of bonds written to pdb")


@pytest.mark.parametrize(
    "bond_orders, ref_orders",
    [
        (False, ["None", None]),
        (True, ["Triple", 3]),
    ],
    ids=["ignore repeated bonds", "read repeated bonds"],
)
def test_bond_order_conect_priority(get_fn, bond_orders, ref_orders):
    """Test that when reading bond orders from pdb, if the bond is listed
    differently in opposite orders, the highest bond order prevails"""

    traj = load_pdb(get_fn("pdb_repeatbond.pdb"), bond_orders=bond_orders)

    bond = [str(traj.top._bonds[0].type), traj.top._bonds[0].order]
    eq(bond, ref_orders, err_msg="Inconsistent bond type, bond order read")



"""
Creates a subclass for all classes intended to be a singleton. This
maintains the correctness of instance is instance even following
pickling/unpickling
"""


class Singleton:
    _inst = None

    def __new__(cls):
        if cls._inst is None:
            cls._inst = super().__new__(cls)
        return cls._inst

    def __reduce__(self):
        return repr(self)

    def __float__(self):
        return 0.0



##
#
# File:     PdbxContainers.py
# Original: 02-Feb-2009   jdw
#
# Update:
#   23-Mar-2011   jdw Added method to rename attributes in category containers.
#   05-Apr-2011   jdw Change cif writer to select double quoting as preferred
#                     quoting style where possible.
#   16-Jan-2012   jdw Create base class for DataCategory class
#   22-Mar-2012   jdw when append attributes to existing categories update
#                     existing rows with placeholder null values.
#    2-Sep-2012   jdw add option to avoid embedded quoting that might
#                     confuse simple parsers.
#   28-Jun-2013   jdw export remove method
#   29-Jun-2013   jdw export remove row method
#   30-Jun-2013   jdw export getRow method
#   15-May-2025   cb changes xrange to range for Python 3 compatibility and uses f-strings.
#                 Adds docstrings, exception handling
##
"""

A collection of container classes supporting the PDBx/mmCIF storage model.

A base container class is defined which supports common features of
data and definition containers.   PDBx data files are organized in
sections called data blocks which are mapped to data containers.
PDBx dictionaries contain definition sections and data sections
which are mapped to definition and data containes respectively.

Data in both PDBx data files and dictionaries are organized in
data categories. In the PDBx syntax individual items or data
identified by labels of the form '_categoryName.attributeName'.
The terms category and attribute in PDBx jargon are analogous
table and column in relational data model, or class and attribute
in an object oriented data model.

The DataCategory class provides base storage container for instance
data and definition meta data.

"""

__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import re
import sys
import traceback


class CifName:
    """
    Utilities for handling CIF-style data names.
    """

    def __init__(self):
        pass

    @staticmethod
    def categoryPart(name):
        """
        Extract the category part from a CIF data name. (e.g. "_atom_site.id" -> "atom_site")

        Parameters
        ----------
        name : str
            The full CIF data name.

        Returns
        -------
        str
            The category portion of the name.
        """
        tname = ""
        if name.startswith("_"):
            tname = name[1:]
        else:
            tname = name

        i = tname.find(".")
        if i == -1:
            return tname
        else:
            return tname[:i]

    @staticmethod
    def attributePart(name):
        """
        Extract the attribute part from a CIF data name. (e.g. "_atom_site.id" -> "id")

        Parameters
        ----------
        name : str
            The full CIF data name.

        Returns
        -------
        str or None
            The attribute portion of the name or None if not found.
        """
        i = name.find(".")
        if i == -1:
            return None
        else:
            return name[i + 1 :]


class ContainerBase:
    """
    Container base class for data and definition objects.

    Parameters
    ----------
    name : str
        The name of the container (e.g. a data block or save frame).
    """

    def __init__(self, name):
        # The enclosing scope of the data container (e.g. data_/save_)
        self.__name = name
        # List of category names within this container -
        self.__objNameList = []
        # dictionary of DataCategory objects keyed by category name.
        self.__objCatalog = {}
        self.__type = None

    def getType(self):
        """
        Get the container type.

        Returns
        -------
        type : any
            The type of the container.
        """
        return self.__type

    def setType(self, type):
        """
        Set the container type.

        Parameters
        ----------
        type : any
            The container type.
        """
        self.__type = type

    def getName(self):
        """
        Get the container name.

        Returns
        -------
        str
            The container name.
        """
        return self.__name

    def setName(self, name):
        """
        Set the container name.

        Parameters
        ----------
        name : str
            The new container name.
        """
        self.__name = name

    def exists(self, name):
        if name in self.__objCatalog:
            return True
        else:
            return False

    def getObj(self, name):
        if name in self.__objCatalog:
            return self.__objCatalog[name]
        else:
            return None

    def getObjNameList(self):
        return self.__objNameList

    def append(self, obj):
        """
        Add the input object to the current object catalog. An existing object
        of the same name will be overwritten.

        Parameters
        ----------
        obj : object
            The object to add. It must have a getName() method.
        """
        if obj.getName() is not None:
            if obj.getName() not in self.__objCatalog:
                # self.__objNameList is keeping track of object order here --
                self.__objNameList.append(obj.getName())
            self.__objCatalog[obj.getName()] = obj

    def replace(self, obj):
        """
        Replace an existing object with the input object.

        Parameters
        ----------
        obj : object
            The object to replace. It must have a getName() method.
        """
        if (obj.getName() is not None) and (obj.getName() in self.__objCatalog):
            self.__objCatalog[obj.getName()] = obj

    def printIt(self, fh=sys.stdout, type="brief"):
        fh.write(
            f"+ {self.getType()} container: {self.getName():>30} contains {len(self.__objNameList):4d} categories\n",
        )
        for nm in self.__objNameList:
            fh.write("--------------------------------------------\n")
            fh.write(f"Data category: {nm}\n")
            if type == "brief":
                self.__objCatalog[nm].printIt(fh)
            else:
                self.__objCatalog[nm].dumpIt(fh)

    def rename(self, curName, newName):
        """
        Change the name of an object in place.

        Parameters
        ----------
        curName : str
            The current name of the object.
        newName : str
            The new name for the object.

        Returns
        -------
        bool
            True if renaming was successful, False otherwise.
        """
        try:
            i = self.__objNameList.index(curName)
            self.__objNameList[i] = newName
            self.__objCatalog[newName] = self.__objCatalog[curName]
            self.__objCatalog[newName].setName(newName)
            return True
        except (KeyError, ValueError):
            return False

    def remove(self, curName):
        """
        Remove an object by name.

        Parameters
        ----------
        curName : str
            The name of the object to remove.

        Returns
        -------
        bool
            True on success, False otherwise.
        """
        try:
            if curName in self.__objCatalog:
                del self.__objCatalog[curName]
                i = self.__objNameList.index(curName)
                del self.__objNameList[i]
                return True
            else:
                return False
        except (KeyError, ValueError):
            pass

        return False


class DefinitionContainer(ContainerBase):
    def __init__(self, name):
        super().__init__(name)
        self.setType("definition")

    def isCategory(self):
        if self.exists("category"):
            return True
        return False

    def isAttribute(self):
        if self.exists("item"):
            return True
        return False

    def printIt(self, fh=sys.stdout, type="brief"):
        fh.write(
            f"Definition container: {self.getName():>30} contains {len(self.getObjNameList()):4d} categories\n",
        )
        if self.isCategory():
            fh.write("Definition type: category\n")
        elif self.isAttribute():
            fh.write("Definition type: item\n")
        else:
            fh.write("Definition type: undefined\n")

        for nm in self.getObjNameList():
            fh.write("--------------------------------------------\n")
            fh.write(f"Definition category: {nm}\n")
            if type == "brief":
                self.getObj(nm).printIt(fh)
            else:
                self.getObj(nm).dumpId(fh)


class DataContainer(ContainerBase):
    """
    Container for DataCategory objects.

    Parameters
    ----------
    name : str
        The name of the data container.
    """

    def __init__(self, name):
        super().__init__(name)
        self.setType("data")
        self.__globalFlag = False

    def invokeDataBlockMethod(self, type, method, db):
        self.__currentRow = 1
        exec(method.getInline())

    def setGlobal(self):
        self.__globalFlag = True

    def getGlobal(self):
        return self.__globalFlag


class DataCategoryBase:
    """
    Base object for a data category.

    Parameters
    ----------
    name : str
        The name of the data category.
    attributeNameList : list of str, optional
        A list of attribute names.
    rowList : list of list, optional
        A list of data rows.
    """

    def __init__(self, name, attributeNameList=None, rowList=None):
        self._name = name
        #
        if rowList is not None:
            self._rowList = rowList
        else:
            self._rowList = []

        if attributeNameList is not None:
            self._attributeNameList = attributeNameList
        else:
            self._attributeNameList = []
        #
        # Derived class data -
        #
        self._catalog = {}
        self._numAttributes = 0
        #
        self.__setup()

    def __setup(self):
        self._numAttributes = len(self._attributeNameList)
        self._catalog = {}
        for attributeName in self._attributeNameList:
            attributeNameLC = attributeName.lower()
            self._catalog[attributeNameLC] = attributeName

    #
    def setRowList(self, rowList):
        self._rowList = rowList

    def setAttributeNameList(self, attributeNameList):
        self._attributeNameList = attributeNameList
        self.__setup()

    def setName(self, name):
        self._name = name

    def get(self):
        return (self._name, self._attributeNameList, self._rowList)


class DataCategory(DataCategoryBase):
    """
    Methods for creating, accessing, and formatting PDBx cif data categories.

    Parameters
    ----------
    name : str
        The name of the data category.
    attributeNameList : list of str, optional
        The list of attribute names.
    rowList : list of list, optional
        The list of data rows.
    """

    def __init__(self, name, attributeNameList=None, rowList=None):
        super().__init__(name, attributeNameList, rowList)
        #
        self.__lfh = sys.stdout

        self.__currentRowIndex = 0
        self.__currentAttribute = None
        #
        self.__avoidEmbeddedQuoting = False
        #
        # --------------------------------------------------------------------
        # any whitespace
        self.__wsRe = re.compile(r"\s")
        self.__wsAndQuotesRe = re.compile(r"[\s'\"]")
        # any newline or carriage control
        self.__nlRe = re.compile(r"[\n\r]")
        #
        # single quote
        self.__sqRe = re.compile(r"[']")
        #
        self.__sqWsRe = re.compile(r"('\s)|(\s')")

        # double quote
        self.__dqRe = re.compile(r'["]')
        self.__dqWsRe = re.compile(r'("\s)|(\s")')
        #
        self.__intRe = re.compile(r"^[0-9]+$")
        self.__floatRe = re.compile(
            r"^-?(([0-9]+)[.]?|([0-9]*[.][0-9]+))([(][0-9]+[)])?([eE][+-]?[0-9]+)?$",
        )
        #
        self.__dataTypeList = [
            "DT_NULL_VALUE",
            "DT_INTEGER",
            "DT_FLOAT",
            "DT_UNQUOTED_STRING",
            "DT_ITEM_NAME",
            "DT_DOUBLE_QUOTED_STRING",
            "DT_SINGLE_QUOTED_STRING",
            "DT_MULTI_LINE_STRING",
        ]
        self.__formatTypeList = [
            "FT_NULL_VALUE",
            "FT_NUMBER",
            "FT_NUMBER",
            "FT_UNQUOTED_STRING",
            "FT_QUOTED_STRING",
            "FT_QUOTED_STRING",
            "FT_QUOTED_STRING",
            "FT_MULTI_LINE_STRING",
        ]
        #

    def __getitem__(self, x):
        """Implements list-type functionality -
             Implements op[x] for some special cases -

        Parameters
        ----------
        x : int or str
            If x is an integer, return the row at that index.
            If x is a string, return the value for that attribute in the first row.

        Returns
        -------
        list or object
            The requested row or attribute value.

        Raises
        ------
        KeyError
            If an attribute is not found.
        TypeError
            If x is neither int nor str.
        """
        if isinstance(x, int):
            # return self._rowList.__getitem__(x)
            return self._rowList[x]

        elif isinstance(x, str):
            try:
                # return self._rowList[0][x]
                ii = self.getAttributeIndex(x)
                return self._rowList[0][ii]
            except (IndexError, KeyError):
                raise KeyError
        raise TypeError(x)

    def getCurrentAttribute(self):
        return self.__currentAttribute

    def getRowIndex(self):
        return self.__currentRowIndex

    def getRowList(self):
        return self._rowList

    def getRowCount(self):
        return len(self._rowList)

    def getRow(self, index):
        try:
            return self._rowList[index]
        except (IndexError, KeyError, ValueError, TypeError):
            return []

    def removeRow(self, index):
        try:
            if (index >= 0) and (index < len(self._rowList)):
                del self._rowList[index]
                if self.__currentRowIndex >= len(self._rowList):
                    self.__currentRowIndex = len(self._rowList) - 1
                return True
            else:
                pass
        except (IndexError, ValueError, TypeError):
            pass

        return False

    def getFullRow(self, index):
        """
        Return a full row based on the length of the the attribute list.

        Parameters
        ----------
        index : int
            The row index.

        Returns
        -------
        list
            The full row with length equal to the number of attributes.
        """
        try:
            if len(self._rowList[index]) < self._numAttributes:
                for ii in range(self._numAttributes - len(self._rowList[index])):
                    self._rowList[index].append("?")
            return self._rowList[index]
        except (IndexError, TypeError):
            return ["?" for ii in range(self._numAttributes)]

    def getName(self):
        return self._name

    def getAttributeList(self):
        return self._attributeNameList

    def getAttributeCount(self):
        return len(self._attributeNameList)

    def getAttributeListWithOrder(self):
        oL = []
        for ii, att in enumerate(self._attributeNameList):
            oL.append((att, ii))
        return oL

    def getAttributeIndex(self, attributeName):
        try:
            return self._attributeNameList.index(attributeName)
        except ValueError:
            return -1

    def hasAttribute(self, attributeName):
        return attributeName in self._attributeNameList

    def getIndex(self, attributeName):
        try:
            return self._attributeNameList.index(attributeName)
        except ValueError:
            return -1

    def getItemNameList(self):
        itemNameList = []
        for att in self._attributeNameList:
            itemNameList.append("_" + self._name + "." + att)
        return itemNameList

    def append(self, row):
        # self.__lfh.write("PdbxContainer(append) category %s row %r\n"  % (self._name,row))
        self._rowList.append(row)

    def appendAttribute(self, attributeName):
        attributeNameLC = attributeName.lower()
        if attributeNameLC in self._catalog:
            i = self._attributeNameList.index(self._catalog[attributeNameLC])
            self._attributeNameList[i] = attributeName
            self._catalog[attributeNameLC] = attributeName
            # self.__lfh.write("Appending existing attribute %s\n" % attributeName)
        else:
            # self.__lfh.write("Appending existing attribute %s\n" % attributeName)
            self._attributeNameList.append(attributeName)
            self._catalog[attributeNameLC] = attributeName
            #
        self._numAttributes = len(self._attributeNameList)

    def appendAttributeExtendRows(self, attributeName):
        attributeNameLC = attributeName.lower()
        if attributeNameLC in self._catalog:
            i = self._attributeNameList.index(self._catalog[attributeNameLC])
            self._attributeNameList[i] = attributeName
            self._catalog[attributeNameLC] = attributeName
            self.__lfh.write(f"Appending existing attribute {attributeName}\n")
        else:
            self._attributeNameList.append(attributeName)
            self._catalog[attributeNameLC] = attributeName
            # add a placeholder to any existing rows for the new attribute.
            if len(self._rowList) > 0:
                for row in self._rowList:
                    row.append("?")
            #
        self._numAttributes = len(self._attributeNameList)

    def getValue(self, attributeName=None, rowIndex=None):
        if attributeName is None:
            attribute = self.__currentAttribute
        else:
            attribute = attributeName
        if rowIndex is None:
            rowI = self.__currentRowIndex
        else:
            rowI = rowIndex

        if isinstance(attribute, str) and isinstance(rowI, int):
            try:
                return self._rowList[rowI][self._attributeNameList.index(attribute)]
            except IndexError:
                raise IndexError
        raise IndexError(attribute)

    def setValue(self, value, attributeName=None, rowIndex=None):
        if attributeName is None:
            attribute = self.__currentAttribute
        else:
            attribute = attributeName

        if rowIndex is None:
            rowI = self.__currentRowIndex
        else:
            rowI = rowIndex

        if isinstance(attribute, str) and isinstance(rowI, int):
            try:
                # if row index is out of range - add the rows -
                for ii in range(rowI + 1 - len(self._rowList)):
                    self._rowList.append(self.__emptyRow())
                # self._rowList[rowI][attribute]=value
                ll = len(self._rowList[rowI])
                ind = self._attributeNameList.index(attribute)

                # extend the list if needed -
                if ind >= ll:
                    self._rowList[rowI].extend([None for ii in range(2 * ind - ll)])
                self._rowList[rowI][ind] = value
            except IndexError:
                self.__lfh.write(
                    f"DataCategory(setvalue) index error "
                    f" category {self._name} attribute {attribute} index {rowI} value {value!r}\n",
                )
                traceback.print_exc(file=self.__lfh)
                # raise IndexError
            except ValueError:
                self.__lfh.write(
                    f"DataCategory(setvalue) value error "
                    f"category {self._name} attribute {attribute} index {rowI} value {value!r}\n",
                )
                traceback.print_exc(file=self.__lfh)
                # raise ValueError

    def __emptyRow(self):
        return [None for ii in range(len(self._attributeNameList))]

    def replaceValue(self, oldValue, newValue, attributeName):
        numReplace = 0
        if attributeName not in self._attributeNameList:
            return numReplace
        ind = self._attributeNameList.index(attributeName)
        for row in self._rowList:
            if row[ind] == oldValue:
                row[ind] = newValue
                numReplace += 1
        return numReplace

    def replaceSubstring(self, oldValue, newValue, attributeName):
        ok = False
        if attributeName not in self._attributeNameList:
            return ok
        ind = self._attributeNameList.index(attributeName)
        for row in self._rowList:
            val = row[ind]
            row[ind] = val.replace(oldValue, newValue)
            if val != row[ind]:
                ok = True
        return ok

    def invokeAttributeMethod(self, attributeName, type, method, db):
        self.__currentRowIndex = 0
        self.__currentAttribute = attributeName
        self.appendAttribute(attributeName)
        ind = self._attributeNameList.index(attributeName)
        if len(self._rowList) == 0:
            row = [None for ii in range(len(self._attributeNameList) * 2)]
            row[ind] = None
            self._rowList.append(row)

        for row in self._rowList:
            ll = len(row)
            if ind >= ll:
                row.extend([None for ii in range(2 * ind - ll)])
                row[ind] = None
            exec(method.getInline())
            self.__currentRowIndex += 1

    def invokeCategoryMethod(self, type, method, db):
        self.__currentRowIndex = 0
        exec(method.getInline())

    def getAttributeLengthMaximumList(self):
        mList = [0 for i in len(self._attributeNameList)]
        for row in self._rowList:
            for indx, val in enumerate(row):
                mList[indx] = max(mList[indx], len(val))
        return mList

    def renameAttribute(self, curAttributeName, newAttributeName):
        """
        Change the name of an attribute in place.

        Parameters
        ----------
        curAttributeName : str
            The current attribute name.
        newAttributeName : str
            The new attribute name.

        Returns
        -------
        bool
            True if renaming was successful, False otherwise.
        """
        try:
            i = self._attributeNameList.index(curAttributeName)
            self._attributeNameList[i] = newAttributeName
            del self._catalog[curAttributeName.lower()]
            self._catalog[newAttributeName.lower()] = newAttributeName
            return True
        except ValueError:
            return False

    def printIt(self, fh=sys.stdout):
        fh.write("--------------------------------------------\n")
        fh.write(
            f"  Category: {self._name} attribute list length: {len(self._attributeNameList)}\n",
        )
        for at in self._attributeNameList:
            fh.write(f"  Category: {self._name} attribute: {at}\n")

        fh.write(f"  Row value list length: {len(self._rowList)}\n")
        #
        for row in self._rowList[:2]:
            #
            if len(row) == len(self._attributeNameList):
                for ii, v in enumerate(row):
                    fh.write(
                        f"        {self._attributeNameList[ii]:30s}: {str(v)[:30]} ...\n",
                    )
            else:
                fh.write(
                    f"+WARNING - {self._name} data length {len(row)} attribute name length "
                    f"{len(self._attributeNameList)} mismatched\n",
                )

    def dumpIt(self, fh=sys.stdout):
        fh.write("--------------------------------------------\n")
        fh.write(
            f"  Category: {self._name} attribute list length: {len(self._attributeNameList)}\n",
        )
        for at in self._attributeNameList:
            fh.write(f"  Category: {self._name} attribute: {at}\n")

        fh.write(f"  Value list length: {len(self._rowList)}\n")
        for row in self._rowList:
            for ii, v in enumerate(row):
                fh.write(f"        {self._attributeNameList[ii]:30s}: {v}\n")

    def __formatPdbx(self, inp):
        """
        Format input data following PDBx quoting rules.

        Parameters
        ----------
        inp : any
            The input data to format.

        Returns
        -------
        tuple
            A tuple (list, type) where the list contains formatted components and
            type is a string representing the data type.
        """
        try:
            if inp is None:
                return ("?", "DT_NULL_VALUE")

            # pure numerical values are returned as unquoted strings
            if isinstance(inp, int) or self.__intRe.search(str(inp)):
                return ([str(inp)], "DT_INTEGER")

            if isinstance(inp, float) or self.__floatRe.search(str(inp)):
                return ([str(inp)], "DT_FLOAT")

            # null value handling -

            if inp == "." or inp == "?":
                return ([inp], "DT_NULL_VALUE")

            if inp == "":
                return (["."], "DT_NULL_VALUE")

            # Contains white space or quotes ?
            if not self.__wsAndQuotesRe.search(inp):
                if inp.startswith("_"):
                    return (self.__doubleQuotedList(inp), "DT_ITEM_NAME")
                else:
                    return ([str(inp)], "DT_UNQUOTED_STRING")
            else:
                if self.__nlRe.search(inp):
                    return (self.__semiColonQuotedList(inp), "DT_MULTI_LINE_STRING")
                else:
                    if self.__avoidEmbeddedQuoting:
                        # change priority to choose double quoting where possible.
                        if not self.__dqRe.search(inp) and not self.__sqWsRe.search(
                            inp,
                        ):
                            return (
                                self.__doubleQuotedList(inp),
                                "DT_DOUBLE_QUOTED_STRING",
                            )
                        elif not self.__sqRe.search(inp) and not self.__dqWsRe.search(
                            inp,
                        ):
                            return (
                                self.__singleQuotedList(inp),
                                "DT_SINGLE_QUOTED_STRING",
                            )
                        else:
                            return (
                                self.__semiColonQuotedList(inp),
                                "DT_MULTI_LINE_STRING",
                            )
                    else:
                        # change priority to choose double quoting where possible.
                        if not self.__dqRe.search(inp):
                            return (
                                self.__doubleQuotedList(inp),
                                "DT_DOUBLE_QUOTED_STRING",
                            )
                        elif not self.__sqRe.search(inp):
                            return (
                                self.__singleQuotedList(inp),
                                "DT_SINGLE_QUOTED_STRING",
                            )
                        else:
                            return (
                                self.__semiColonQuotedList(inp),
                                "DT_MULTI_LINE_STRING",
                            )

        except Exception:
            traceback.print_exc(file=self.__lfh)

    def __dataTypePdbx(self, inp):
        """
        Detect the PDBx data type of the input.

        Parameters
        ----------
        inp : any
            The input data.

        Returns
        -------
        str
            The detected data type.
        """
        if inp is None:
            return "DT_NULL_VALUE"

        # pure numerical values are returned as unquoted strings
        if isinstance(inp, int) or self.__intRe.search(str(inp)):
            return "DT_INTEGER"

        if isinstance(inp, float) or self.__floatRe.search(str(inp)):
            return "DT_FLOAT"

        # null value handling -

        if inp == "." or inp == "?":
            return "DT_NULL_VALUE"

        if inp == "":
            return "DT_NULL_VALUE"

        # Contains white space or quotes ?
        if not self.__wsAndQuotesRe.search(inp):
            if inp.startswith("_"):
                return "DT_ITEM_NAME"
            else:
                return "DT_UNQUOTED_STRING"
        else:
            if self.__nlRe.search(inp):
                return "DT_MULTI_LINE_STRING"
            else:
                if self.__avoidEmbeddedQuoting:
                    if not self.__sqRe.search(inp) and not self.__dqWsRe.search(inp):
                        return "DT_DOUBLE_QUOTED_STRING"
                    elif not self.__dqRe.search(inp) and not self.__sqWsRe.search(inp):
                        return "DT_SINGLE_QUOTED_STRING"
                    else:
                        return "DT_MULTI_LINE_STRING"
                else:
                    if not self.__sqRe.search(inp):
                        return "DT_DOUBLE_QUOTED_STRING"
                    elif not self.__dqRe.search(inp):
                        return "DT_SINGLE_QUOTED_STRING"
                    else:
                        return "DT_MULTI_LINE_STRING"

    def __singleQuotedList(self, inp):
        quoted_list = []
        quoted_list.append("'")
        quoted_list.append(inp)
        quoted_list.append("'")
        return quoted_list

    def __doubleQuotedList(self, inp):
        quoted_list = []
        quoted_list.append('"')
        quoted_list.append(inp)
        quoted_list.append('"')
        return quoted_list

    def __semiColonQuotedList(self, inp):
        quoted_list = []
        quoted_list.append("\n")
        if inp[-1] == "\n":
            quoted_list.append(";")
            quoted_list.append(inp)
            quoted_list.append(";")
            quoted_list.append("\n")
        else:
            quoted_list.append(";")
            quoted_list.append(inp)
            quoted_list.append("\n")
            quoted_list.append(";")
            quoted_list.append("\n")

        return quoted_list

    def getValueFormatted(self, attributeName=None, rowIndex=None):
        if attributeName is None:
            attribute = self.__currentAttribute
        else:
            attribute = attributeName

        if rowIndex is None:
            rowI = self.__currentRowIndex
        else:
            rowI = rowIndex

        if isinstance(attribute, str) and isinstance(rowI, int):
            try:
                list_, type_ = self.__formatPdbx(
                    self._rowList[rowI][self._attributeNameList.index(attribute)],
                )
                return "".join(list_)
            except IndexError:
                self.__lfh.write(
                    f"attributeName {attributeName} rowI {rowI!r} rowdata {self._rowList[rowI]!r}\n",
                )
                raise IndexError
        raise TypeError(attribute)

    def getValueFormattedByIndex(self, attributeIndex, rowIndex):
        try:
            list_, type_ = self.__formatPdbx(self._rowList[rowIndex][attributeIndex])
            return "".join(list_)
        except IndexError:
            raise IndexError

    def getAttributeValueMaxLengthList(self, steps=1):
        mList = [0 for i in range(len(self._attributeNameList))]
        for row in self._rowList[::steps]:
            for indx in range(len(self._attributeNameList)):
                val = row[indx]
                mList[indx] = max(mList[indx], len(str(val)))
        return mList

    def getFormatTypeList(self, steps=1):
        try:
            curDataTypeList = ["DT_NULL_VALUE" for i in range(len(self._attributeNameList))]
            for row in self._rowList[::steps]:
                for indx in range(len(self._attributeNameList)):
                    val = row[indx]
                    dType = self.__dataTypePdbx(val)
                    dIndx = self.__dataTypeList.index(dType)
                    cType = curDataTypeList[indx]
                    cIndx = self.__dataTypeList.index(cType)
                    cIndx = max(cIndx, dIndx)
                    curDataTypeList[indx] = self.__dataTypeList[cIndx]

            curFormatTypeList = []
            for dt in curDataTypeList:
                ii = self.__dataTypeList.index(dt)
                curFormatTypeList.append(self.__formatTypeList[ii])
        except Exception:
            self.__lfh.write(
                f"PdbxDataCategory(getFormatTypeList) ++Index error at index {indx} in row {row}\n",
            )

        return curFormatTypeList, curDataTypeList

    def getFormatTypeListX(self):
        curDataTypeList = ["DT_NULL_VALUE" for i in range(len(self._attributeNameList))]
        for row in self._rowList:
            for indx in range(len(self._attributeNameList)):
                val = row[indx]
                # print("index ",indx," val ",val)
                dType = self.__dataTypePdbx(val)
                dIndx = self.__dataTypeList.index(dType)
                # print("d type", dType, " d type index ",dIndx)

                cType = curDataTypeList[indx]
                cIndx = self.__dataTypeList.index(cType)
                cIndx = max(cIndx, dIndx)
                curDataTypeList[indx] = self.__dataTypeList[cIndx]

        # Map the format types to the data types
        curFormatTypeList = []
        for dt in curDataTypeList:
            ii = self.__dataTypeList.index(dt)
            curFormatTypeList.append(self.__formatTypeList[ii])
        return curFormatTypeList, curDataTypeList



"""MDTraj: A modern, open library for the analysis of molecular dynamics trajectories

MDTraj is a python library that allows users to manipulate molecular dynamics
(MD) trajectories and perform a variety of analyses, including fast RMSD,
solvent accessible surface area, hydrogen bonding, etc. A highlight of MDTraj
is the wide variety of molecular dynamics trajectory file formats which are
supported, including RCSB pdb, GROMACS xtc, and trr, CHARMM / NAMD dcd, AMBER
AMBER NetCDF, AMBER mdcrd, TINKER arc and MDTraj HDF5.
"""

import sys

import versioneer
from setuptools import Extension, find_packages, setup

from basesetup import (
    CompilerDetection,
    StaticLibrary,
    build_ext,
    parse_setuppy_commands,
)

DOCLINES = __doc__.split("\n")

sys.path.insert(0, ".")

try:
    # add an optional --disable-openmp to disable OpenMP support
    sys.argv.remove("--disable-openmp")
    disable_openmp = True
except ValueError:
    disable_openmp = False


CLASSIFIERS = """\
Development Status :: 5 - Production/Stable
Intended Audience :: Science/Research
Intended Audience :: Developers
License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)
Programming Language :: C
Programming Language :: Python
Programming Language :: Python :: 3
Topic :: Scientific/Engineering :: Bio-Informatics
Topic :: Scientific/Engineering :: Chemistry
Operating System :: Microsoft :: Windows
Operating System :: POSIX
Operating System :: Unix
Operating System :: MacOS
"""

# Global info about compiler
compiler = CompilerDetection(disable_openmp)
compiler.initialize()

extra_cpp_libraries = []

if sys.platform == "win32":
    extra_cpp_libraries.append("Ws2_32")
    # For determining if a path is relative (for dtr)
    extra_cpp_libraries.append("Shlwapi")


################################################################################
# Declaration of the compiled extension modules (cython + c)
################################################################################


def format_extensions():
    compiler_args = compiler.compiler_args_warn

    xtc = Extension(
        "mdtraj.formats.xtc",
        sources=[
            "mdtraj/formats/xtc/src/xdrfile.c",
            "mdtraj/formats/xtc/src/xdr_seek.c",
            "mdtraj/formats/xtc/src/xdrfile_xtc.c",
            "mdtraj/formats/xtc/xtc.pyx",
        ],
        include_dirs=[
            "mdtraj/formats/xtc/include/",
            "mdtraj/formats/xtc/",
        ],
        extra_compile_args=compiler_args,
    )

    trr = Extension(
        "mdtraj.formats.trr",
        sources=[
            "mdtraj/formats/xtc/src/xdrfile.c",
            "mdtraj/formats/xtc/src/xdr_seek.c",
            "mdtraj/formats/xtc/src/xdrfile_trr.c",
            "mdtraj/formats/xtc/trr.pyx",
        ],
        include_dirs=[
            "mdtraj/formats/xtc/include/",
            "mdtraj/formats/xtc/",
        ],
        extra_compile_args=compiler_args,
    )

    zlib_include_dirs = []
    zlib_library_dirs = []
    if sys.platform == "win32":
        # Conda puts the zlib headers in ./Library/... on windows
        # If you're not using conda, good luck!
        zlib_include_dirs += [f"{sys.prefix}/Library/include"]
        zlib_library_dirs += [f"{sys.prefix}/Library/lib"]
    else:
        # On linux (and mac(?)) these paths should work for a standard
        # install of python+zlib or a conda install of python+zlib
        zlib_include_dirs += [f"{sys.prefix}/include"]
        zlib_library_dirs += [f"{sys.prefix}/lib"]

    dcd = Extension(
        "mdtraj.formats.dcd",
        sources=[
            "mdtraj/formats/dcd/src/dcdplugin.c",
            "mdtraj/formats/dcd/dcd.pyx",
        ],
        include_dirs=[
            "mdtraj/formats/dcd/include/",
            "mdtraj/formats/dcd/",
        ],
        extra_compile_args=compiler_args,
    )

    dtr = Extension(
        "mdtraj.formats.dtr",
        sources=[
            "mdtraj/formats/dtr/src/dtrplugin.cxx",
            "mdtraj/formats/dtr/dtr.pyx",
        ],
        include_dirs=[
            "mdtraj/formats/dtr/include/",
            "mdtraj/formats/dtr/",
        ],
        define_macros=[("DESRES_READ_TIMESTEP2", 1)],
        language="c++",
        extra_compile_args=compiler_args,
        libraries=extra_cpp_libraries,
    )

    return [xtc, trr, dcd, dtr]


def rmsd_extensions():
    compiler_args = (
        compiler.compiler_args_openmp
        + compiler.compiler_args_sse2
        + compiler.compiler_args_sse3
        + compiler.compiler_args_opt
        + compiler.compiler_args_warn
    )
    compiler_libraries = compiler.compiler_libraries_openmp
    define_macros = [("__NO_INTRINSICS", 1)] if compiler.disable_intrinsics else None

    libtheobald = StaticLibrary(
        "mdtraj.core.lib.libtheobald",
        sources=[
            "mdtraj/rmsd/src/theobald_rmsd.cpp",
            "mdtraj/rmsd/src/center.cpp",
        ],
        include_dirs=[
            "mdtraj/rmsd/include",
        ],
        export_include=[
            "mdtraj/rmsd/include/theobald_rmsd.h",
            "mdtraj/rmsd/include/center.h",
        ],
        language="c++",
        # don't enable OpenMP
        extra_compile_args=(compiler.compiler_args_sse2 + compiler.compiler_args_sse3 + compiler.compiler_args_opt),
    )

    rmsd = Extension(
        "mdtraj._rmsd",
        sources=[
            "mdtraj/rmsd/src/theobald_rmsd.cpp",
            "mdtraj/rmsd/src/rotation.cpp",
            "mdtraj/rmsd/src/center.cpp",
            "mdtraj/rmsd/_rmsd.pyx",
        ],
        include_dirs=["mdtraj/rmsd/include"],
        define_macros=define_macros,
        extra_compile_args=compiler_args,
        libraries=compiler_libraries,
        language="c++",
    )

    lprmsd = Extension(
        "mdtraj._lprmsd",
        sources=[
            "mdtraj/rmsd/src/theobald_rmsd.cpp",
            "mdtraj/rmsd/src/rotation.cpp",
            "mdtraj/rmsd/src/center.cpp",
            "mdtraj/rmsd/src/fancy_index.cpp",
            "mdtraj/rmsd/src/Munkres.cpp",
            "mdtraj/rmsd/src/euclidean_permutation.cpp",
            "mdtraj/rmsd/_lprmsd.pyx",
        ],
        language="c++",
        define_macros=define_macros,
        include_dirs=["mdtraj/rmsd/include"],
        extra_compile_args=compiler_args,
        libraries=compiler_libraries + extra_cpp_libraries,
    )
    return rmsd, lprmsd, libtheobald


def geometry_extensions():
    compiler.initialize()
    compiler_args = (
        compiler.compiler_args_openmp
        + compiler.compiler_args_sse2
        + compiler.compiler_args_sse3
        + compiler.compiler_args_opt
        + compiler.compiler_args_warn
    )
    define_macros = [("__NO_INTRINSICS", 1)] if compiler.disable_intrinsics else None
    compiler_libraries = compiler.compiler_libraries_openmp + extra_cpp_libraries

    return [
        Extension(
            "mdtraj.geometry._geometry",
            sources=[
                "mdtraj/geometry/src/sasa.cpp",
                "mdtraj/geometry/src/dssp.cpp",
                "mdtraj/geometry/src/geometry.cpp",
                "mdtraj/geometry/src/_geometry.pyx",
            ],
            include_dirs=[
                "mdtraj/geometry/include",
                "mdtraj/geometry/src/kernels",
            ],
            depends=[
                "mdtraj/geometry/src/kernels/anglekernels.h",
                "mdtraj/geometry/src/kernels/dihedralkernels.h",
                "mdtraj/geometry/src/kernels/distancekernels.h",
            ],
            define_macros=define_macros,
            extra_compile_args=compiler_args,
            libraries=compiler_libraries,
            language="c++",
        ),
        Extension(
            "mdtraj.geometry.drid",
            sources=[
                "mdtraj/geometry/drid.pyx",
                "mdtraj/geometry/src/dridkernels.cpp",
                "mdtraj/geometry/src/moments.cpp",
            ],
            include_dirs=["mdtraj/geometry/include"],
            define_macros=define_macros,
            extra_compile_args=compiler_args,
            libraries=compiler_libraries,
            language="c++",
        ),
        Extension(
            "mdtraj.geometry.neighbors",
            sources=[
                "mdtraj/geometry/neighbors.pyx",
                "mdtraj/geometry/src/neighbors.cpp",
            ],
            include_dirs=["mdtraj/geometry/include"],
            define_macros=define_macros,
            extra_compile_args=compiler_args,
            libraries=compiler_libraries,
            language="c++",
        ),
        Extension(
            "mdtraj.geometry.neighborlist",
            sources=[
                "mdtraj/geometry/neighborlist.pyx",
                "mdtraj/geometry/src/neighborlist.cpp",
            ],
            include_dirs=["mdtraj/geometry/include"],
            define_macros=define_macros,
            extra_compile_args=compiler_args,
            libraries=compiler_libraries,
            language="c++",
        ),
    ]


EXTRAS_REQUIRE = {
    "tests": [
        "pytest",
        "pytest-rerunfailures",
        "pytest-xdist",
    ],
    "optional": [
        "tables",
        "networkx",
        "netCDF4",
        "pandas<3",
    ],
}


metadata = dict(
    name="mdtraj",
    author="Robert McGibbon",
    author_email="rmcgibbo@gmail.com",
    description=DOCLINES[0],
    long_description="\n".join(DOCLINES[2:]),
    version=versioneer.get_version(),
    license="LGPLv2.1+",
    url="http://mdtraj.org",
    download_url="https://github.com/rmcgibbo/mdtraj/releases/latest",
    platforms=["Linux", "Mac OS-X", "Unix", "Windows"],
    python_requires=">=3.11",
    classifiers=CLASSIFIERS.splitlines(),
    packages=find_packages(),
    cmdclass={**versioneer.get_cmdclass(), "build_ext": build_ext},
    install_requires=[
        "numpy~=2.0",
        "scipy",
        "pyparsing",
        "packaging",
    ],
    extras_require={
        "tests": EXTRAS_REQUIRE["tests"],
        "optional": EXTRAS_REQUIRE["optional"],
        "all": EXTRAS_REQUIRE["tests"] + EXTRAS_REQUIRE["optional"],
    },
    package_data={"mdtraj.formats.pdb": ["data/*"]},
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "mdconvert = mdtraj.scripts.mdconvert:entry_point",
            "mdinspect = mdtraj.scripts.mdinspect:entry_point",
        ],
    },
)


if __name__ == "__main__":
    # Don't use numpy if we are just - non-build actions are required to succeed
    # without NumPy for example when pip is used to install Scipy when
    # NumPy is not yet present in the system.
    run_build = parse_setuppy_commands()
    if run_build:
        extensions = format_extensions()
        extensions.extend(rmsd_extensions())
        extensions.extend(geometry_extensions())

        # most extensions use numpy, add headers for it.
        try:
            import Cython as _c
            from Cython.Build import cythonize

            if _c.__version__ < "0.29":
                raise ImportError("Too old")
        except ImportError as e:
            print(
                "mdtrajs setup depends on Cython (>=0.29). Install it prior invoking setup.py",
            )
            print(e)
            sys.exit(1)
        try:
            import numpy as np
        except ImportError:
            print("mdtrajs setup depends on NumPy. Install it prior invoking setup.py")
            sys.exit(1)

        for e in extensions:
            e.include_dirs.append(np.get_include())
        metadata["ext_modules"] = cythonize(
            extensions,
            language_level=sys.version_info[0],
            force=True,
        )

    setup(**metadata)



from mdtraj import load
from mdtraj.testing import eq


def test_load_single(get_fn):
    # Just check for any raised errors coming from loading a single file.
    load(get_fn("frame0.pdb"))


def test_load_single_top_none(get_fn):
    # Just check for any raised errors coming from loading a single file with top=None
    load(get_fn("frame0.pdb"), top=None)


def test_load_single_list(get_fn):
    # See if a single-element list of files is successfully loaded.
    load([get_fn("frame0.pdb")])


def test_load_many_list(get_fn):
    # See if a multi-element list of files is successfully loaded.
    single = load(get_fn("frame0.pdb"))
    double = load(2 * [get_fn("frame0.pdb")], discard_overlapping_frames=False)
    assert 2 * single.n_frames == double.n_frames


def test_load_atom_indices_multiple_files(get_fn):
    ref_t = load(get_fn("native.pdb"))
    t = load([get_fn("native.pdb")] * 2, atom_indices=[0])

    eq(t.topology, ref_t.topology.subset([0]))



import bz2
import gzip
import io
import os
from io import StringIO


def open_maybe_zipped(filename, mode, force_overwrite=True):
    """Open a file in text (not binary) mode, transparently handling
    .gz or .bz2 compresssion, with utf-8 encoding.

    Parameters
    ----------
    filename : str
        Path to file. Compression will be automatically detected if
        the filename ends in .gz or .bz2.
    mode : {'r', 'w'}
        Mode in which to open file
    force_overwrite : bool, default=True
        If 'w', should we overwrite the file if something with `filename`
        already exists?

    Returns
    -------
    handle : file
        Open file handle.
    """
    _, extension = os.path.splitext(str(filename).lower())
    if mode == "r":
        if extension == ".gz":
            with gzip.GzipFile(filename, "r") as gz_f:
                return StringIO(gz_f.read().decode("utf-8"))
        elif extension == ".bz2":
            with bz2.BZ2File(filename, "r") as bz2_f:
                return StringIO(bz2_f.read().decode("utf-8"))
        else:
            return open(filename)
    elif mode == "w":
        if os.path.exists(filename) and not force_overwrite:
            raise OSError('"%s" already exists' % filename)
        if extension == ".gz":
            binary_fh = gzip.GzipFile(filename, "wb")
            return io.TextIOWrapper(binary_fh, encoding="utf-8")
        elif extension == ".bz2":
            binary_fh = bz2.BZ2File(filename, "wb")
            return io.TextIOWrapper(binary_fh, encoding="utf-8")
        else:
            return open(filename, "w")
    else:
        raise ValueError('Invalid mode "%s"' % mode)



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


"""
This module provides the ability to read and write AMBER NetCDF trajectories.

The code is heavily based on amber_netcdf_trajectory_tools.py by John Chodera.
"""

import os
import socket
import warnings
from datetime import datetime

import numpy as np

import mdtraj
from mdtraj.formats.registry import FormatRegistry
from mdtraj.utils import cast_indices, ensure_type, in_units_of

__all__ = ["NetCDFTrajectoryFile", "load_netcdf"]

##############################################################################
# classes
##############################################################################


@FormatRegistry.register_loader(".nc")
@FormatRegistry.register_loader(".netcdf")
@FormatRegistry.register_loader(".ncdf")
def load_netcdf(filename, top=None, stride=None, atom_indices=None, frame=None):
    """Load an AMBER NetCDF file. Since the NetCDF format doesn't contain
    information to specify the topology, you need to supply a topology

    Parameters
    ----------
    filename : path-like
        filename of AMBER NetCDF file.
    top : {str, Trajectory, Topology}
        The NetCDF format does not contain topology information. Pass in either
        the path to a pdb file, a trajectory, or a topology to supply this
        information.
    stride : int, default=None
        Only read every stride-th frame
    atom_indices : array_like, optional
        If not None, then read only a subset of the atoms coordinates from the
        file. This may be slightly slower than the standard read because it
        requires an extra copy, but will save memory.
    frame : int, optional
        Use this option to load only a single frame from a trajectory on disk.
        If frame is None, the default, the entire trajectory will be loaded.
        If supplied, ``stride`` will be ignored.

    Returns
    -------
    trajectory : md.Trajectory
        The resulting trajectory, as an md.Trajectory object.

    See Also
    --------
    mdtraj.NetCDFTrajectoryFile :  Low level interface to NetCDF files
    """
    from mdtraj.core.trajectory import _parse_topology

    if top is None:
        raise ValueError('"top" argument is required for load_netcdf')

    topology = _parse_topology(top)
    atom_indices = cast_indices(atom_indices)

    with NetCDFTrajectoryFile(filename) as f:
        if frame is not None:
            f.seek(frame)
            n_frames = 1
        else:
            n_frames = None

        return f.read_as_traj(
            topology,
            n_frames=n_frames,
            atom_indices=atom_indices,
            stride=stride,
        )


@FormatRegistry.register_fileobject(".nc")
@FormatRegistry.register_fileobject(".netcdf")
@FormatRegistry.register_fileobject(".ncdf")
class NetCDFTrajectoryFile:
    """Interface for reading and writing to AMBER NetCDF files. This is a
    file-like object, that supports both reading or writing depending
    on the `mode` flag. It implements the context manager protocol,
    so you can also use it with the python 'with' statement.

    Parameters
    ----------
    filename : path-like
        The name of the file to open
    mode : {'r', 'w'}, default='r'
        The mode in which to open the file. Valid options are 'r' and 'w' for
        'read' and 'write', respectively.
    force_overwrite : bool, default=False
        In write mode, if a file named `filename` already exists, clobber
        it and overwrite it.
    """

    distance_unit = "angstroms"

    def __init__(self, filename, mode="r", force_overwrite=True):
        self._closed = True  # is the file currently closed?
        self._mode = mode  # what mode were we opened in

        try:
            # import netcdf4 if it's available
            import netCDF4

            netcdf = netCDF4.Dataset

            # set input args for netCDF4
            input_args = {"format": "NETCDF3_64BIT", "clobber": force_overwrite}

        except ImportError:
            import scipy.io

            netcdf = scipy.io.netcdf_file

            warning_message = (
                "Warning: The 'netCDF4' Python package is not installed. MDTraj is using the 'scipy' "
                "implementation to read and write netCDF files,which can be significantly slower.\n"
                "For improved performance, consider installing netCDF4. See installation instructions at:\n"
                "https://unidata.github.io/netcdf4-python/#quick-install"
            )

            warnings.warn(warning_message)

            # input args for scipy.io.netcdf_file
            # AMBER uses the NetCDF3 format, with 64 bit encodings, which
            # for scipy.io.netcdf_file is "version=2"
            input_args = {"version": 2}

        if mode not in ["r", "w"]:
            raise ValueError("mode must be one of ['r', 'w']")

        if mode == "w" and not force_overwrite and os.path.exists(filename):
            raise OSError('"%s" already exists' % filename)

        self._handle = netcdf(filename, mode=mode, **input_args)
        self._closed = False

        # self._frame_index is the current frame that we're at in the
        #     file
        # self._needs_initialization indicates whether we need to set the
        #     global properties of the file. This is required before the first
        #     write operation on a new file

        if mode == "w":
            self._frame_index = 0
            self._needs_initialization = True
        elif mode == "r":
            self._frame_index = 0
            self._needs_initialization = False
        else:
            raise RuntimeError()

    @property
    def n_atoms(self):
        self._validate_open()
        if self._needs_initialization:
            raise OSError("The file is uninitialized.")

        try:
            # netCDF4 requires use of .size on dimensions
            # to get size as integer.
            return self._handle.dimensions["atom"].size
        except AttributeError:
            # scipy case
            return self._handle.dimensions["atom"]

    @property
    def n_frames(self):
        self._validate_open()
        if not self._needs_initialization:
            return self._handle.variables["coordinates"].shape[0]
        return 0

    def _validate_open(self):
        if self._closed:
            raise OSError("The file is closed.")

    def read_as_traj(self, topology, n_frames=None, stride=None, atom_indices=None):
        """Read a trajectory from a NetCDF file

        Parameters
        ----------
        topology : Topology
            The system topology
        n_frames : int, optional
            If positive, then read only the next `n_frames` frames. Otherwise read all
            of the frames in the file.
        stride : np.ndarray, optional
            Read only every stride-th frame.
        atom_indices : array_like, optional
            If not none, then read only a subset of the atoms coordinates from the
            file. This may be slightly slower than the standard read because it required
            an extra copy, but will save memory.

        Returns
        -------
        trajectory : Trajectory
            A trajectory object containing the loaded portion of the file.
        """
        from mdtraj.core.trajectory import Trajectory

        if atom_indices is not None:
            topology = topology.subset(atom_indices)

        xyz, time, cell_lengths, cell_angles = self.read(
            n_frames=n_frames,
            stride=stride,
            atom_indices=atom_indices,
        )
        if len(xyz) == 0:
            return Trajectory(xyz=np.zeros((0, topology.n_atoms, 3)), topology=topology)

        xyz = in_units_of(
            xyz,
            self.distance_unit,
            Trajectory._distance_unit,
            inplace=True,
        )
        cell_lengths = in_units_of(
            cell_lengths,
            self.distance_unit,
            Trajectory._distance_unit,
            inplace=True,
        )

        return Trajectory(
            xyz=xyz,
            topology=topology,
            time=time,
            unitcell_lengths=cell_lengths,
            unitcell_angles=cell_angles,
        )

    def read(self, n_frames=None, stride=None, atom_indices=None):
        """Read data from a molecular dynamics trajectory in the AMBER NetCDF
        format.

        Parameters
        ----------
        n_frames : int, optional
            If n_frames is not None, the next n_frames of data from the file
            will be read. Otherwise, all of the frames in the file will be read.
        stride : int, optional
            If stride is not None, read only every stride-th frame from disk.
        atom_indices : np.ndarray, dtype=int, optional
            The specific indices of the atoms you'd like to retrieve. If not
            supplied, all of the atoms will be retrieved.

        Returns
        -------
        coordinates : np.ndarray, shape=(n_frames, n_atoms, 3)
            The cartesian coordinates of the atoms, in units of angstroms.
        time : np.ndarray, None
            The time corresponding to each frame, in units of picoseconds, or
            None if no time information is present in the trajectory.
        cell_lengths : np.ndarray, None
            The lengths (a,b,c) of the unit cell for each frame, or None if
            the information is not present in the file.
        cell_angles : np.ndarray, None
            The angles (\alpha, \beta, \\gamma) defining the unit cell for
            each frame, or None if  the information is not present in the file.
        """
        self._validate_open()
        if self._mode != "r":
            raise OSError(
                "The file was opened in mode=%s. Reading is not allowed." % self._mode,
            )

        if n_frames is None:
            n_frames = np.inf
        elif stride is not None:
            # 'n_frames' frames should be read in total
            n_frames *= stride

        total_n_frames = self.n_frames
        frame_slice = slice(
            self._frame_index,
            self._frame_index + min(n_frames, total_n_frames),
            stride,
        )
        if self._frame_index >= total_n_frames:
            # just return something that'll look like len(xyz) == 0
            # this is basically just an alternative to throwing an indexerror
            return np.array([]), None, None, None

        if atom_indices is None:
            # get all of the atoms
            atom_slice = slice(None)
        else:
            atom_slice = ensure_type(
                atom_indices,
                dtype=int,
                ndim=1,
                name="atom_indices",
                warn_on_cast=False,
            )
            if not np.all(atom_slice < self.n_atoms):
                raise ValueError(
                    "As a zero-based index, the entries in "
                    "atom_indices must all be less than the number of atoms "
                    "in the trajectory, %d" % self.n_atoms,
                )
            if not np.all(atom_slice >= 0):
                raise ValueError(
                    "The entries in atom_indices must be greater than or equal to zero",
                )

        if "coordinates" in self._handle.variables:
            coordinates = self._handle.variables["coordinates"][
                frame_slice,
                atom_slice,
                :,
            ]
        else:
            raise ValueError(
                "No coordinates found in the NetCDF file. The only "
                "variables in the file were %s" % self._handle.variables.keys(),
            )

        if "time" in self._handle.variables:
            time = self._handle.variables["time"][frame_slice]
        else:
            time = None

        if "cell_lengths" in self._handle.variables:
            cell_lengths = self._handle.variables["cell_lengths"][frame_slice]
        else:
            cell_lengths = None

        if "cell_angles" in self._handle.variables:
            cell_angles = self._handle.variables["cell_angles"][frame_slice]
        else:
            cell_angles = None

        if cell_lengths is None and cell_angles is not None:
            warnings.warn("cell_lengths were found, but no cell_angles")
        if cell_lengths is not None and cell_angles is None:
            warnings.warn("cell_angles were found, but no cell_lengths")

        self._frame_index = self._frame_index + min(n_frames, total_n_frames)

        # scipy.io.netcdf variables are mem-mapped, and are only backed
        # by valid memory while the file handle is open. This is _bad_.
        # because we need to support the user opening the file, reading
        # the coordinates, and then closing it, and still having the
        # coordinates be a valid memory segment.
        # https://github.com/rmcgibbo/mdtraj/issues/440
        if coordinates is not None and not coordinates.flags["WRITEABLE"]:
            coordinates = np.array(coordinates, copy=True)
        if time is not None and not time.flags["WRITEABLE"]:
            time = np.array(time, copy=True)
        if cell_lengths is not None and not cell_lengths.flags["WRITEABLE"]:
            cell_lengths = np.array(cell_lengths, copy=True)
        if cell_angles is not None and not cell_angles.flags["WRITEABLE"]:
            cell_angles = np.array(cell_angles, copy=True)

        return coordinates, time, cell_lengths, cell_angles

    def write(self, coordinates, time=None, cell_lengths=None, cell_angles=None):
        """Write one or more frames of a molecular dynamics trajectory to disk
        in the AMBER NetCDF format.

        Parameters
        ----------
        coordinates : np.ndarray, dtype=np.float32, shape=(n_frames, n_atoms, 3)
            The cartesian coordinates of each atom, in units of angstroms.
        time : np.ndarray, dtype=np.float32, shape=(n_frames), optional
            The time index corresponding to each frame, in units of picoseconds.
        cell_lengths : np.ndarray, dtype=np.double, shape=(n_frames, 3)
            The lengths (a,b,c) of the unit cell for each frame.
        cell_angles : np.ndarray, dtype=np.double, shape=(n_frames, 3)
            The angles (\alpha, \beta, \\gamma) defining the unit cell for
            each frame.

        Notes
        -----
        If the input arrays are of dimension deficient by one, for example
        if the coordinates array is two dimensional, the time is a single
        scalar or cell_lengths and cell_angles are a 1d array of length three,
        that is okay. You'll simply be saving a single frame.
        """
        self._validate_open()
        if self._mode not in ["w", "ws", "a", "as"]:
            raise OSError(
                "The file was opened in mode=%s. Writing is not allowed." % self._mode,
            )

        coordinates = in_units_of(coordinates, None, "angstroms")
        time = in_units_of(time, None, "picoseconds")
        cell_lengths = in_units_of(cell_lengths, None, "angstroms")
        cell_angles = in_units_of(cell_angles, None, "degrees")

        # typecheck all of the input arguments rigorously
        coordinates = ensure_type(
            coordinates,
            np.float32,
            3,
            "coordinates",
            length=None,
            can_be_none=False,
            shape=(None, None, 3),
            warn_on_cast=False,
            add_newaxis_on_deficient_ndim=True,
        )
        n_frames, n_atoms = coordinates.shape[0], coordinates.shape[1]

        time = ensure_type(
            time,
            np.float32,
            1,
            "time",
            length=n_frames,
            can_be_none=True,
            warn_on_cast=False,
            add_newaxis_on_deficient_ndim=True,
        )
        cell_lengths = ensure_type(
            cell_lengths,
            np.float64,
            2,
            "cell_lengths",
            length=n_frames,
            can_be_none=True,
            shape=(n_frames, 3),
            warn_on_cast=False,
            add_newaxis_on_deficient_ndim=True,
        )
        cell_angles = ensure_type(
            cell_angles,
            np.float64,
            2,
            "cell_angles",
            length=n_frames,
            can_be_none=True,
            shape=(n_frames, 3),
            warn_on_cast=False,
            add_newaxis_on_deficient_ndim=True,
        )

        # are we dealing with a periodic system?
        if (cell_lengths is None and cell_angles is not None) or (cell_lengths is not None and cell_angles is None):
            provided, neglected = "cell_lengths", "cell_angles"
            if cell_lengths is None:
                provided, neglected = neglected, provided
            raise ValueError(
                f'You provided the variable "{provided}", but neglected to '
                f'provide "{neglected}". They either BOTH must be provided, or '
                "neither. Having one without the other is meaningless",
            )

        if self._needs_initialization:
            self._initialize_headers(
                n_atoms=n_atoms,
                set_coordinates=True,
                set_time=(time is not None),
                set_cell=(cell_lengths is not None and cell_angles is not None),
            )
            self._needs_initialization = False

        # this slice object says where we're going to put the data in the
        # arrays
        frame_slice = slice(self._frame_index, self._frame_index + n_frames)

        # deposit the data
        try:
            self._handle.variables["coordinates"][frame_slice, :, :] = coordinates
            if time is not None:
                self._handle.variables["time"][frame_slice] = time
            if cell_lengths is not None:
                self._handle.variables["cell_lengths"][frame_slice, :] = cell_lengths
            if cell_angles is not None:
                self._handle.variables["cell_angles"][frame_slice, :] = cell_angles
        except KeyError as e:
            raise ValueError(
                "The file that you're trying to save to doesn't contain the field %s." % str(e),
            )

        # check for missing attributes
        missing = None
        if time is None and "time" in self._handle.variables:
            missing = "time"
        elif cell_angles is None and "cell_angles" in self._handle.variables:
            missing = "cell_angles"
        elif cell_lengths is None and "cell_lengths" in self._handle.variables:
            missing = "cell_lengths"
        if missing is not None:
            raise ValueError(
                "The file that you're saving to expects each frame "
                "to contain %s information, but you did not supply it."
                "I don't allow 'ragged' arrays." % missing,
            )

        # update the frame index pointers. this should be done at the
        # end so that if anything errors out, we don't actually get here
        self._frame_index += n_frames

    def flush(self):
        "Write all buffered data in the to the disk file."
        self._validate_open()
        self._handle.sync()

    def _initialize_headers(self, set_coordinates, n_atoms, set_time, set_cell):
        """Initialize the NetCDF file according to the AMBER NetCDF Convention,
        Version 1.0, revision B.

        The convention is defined here: http://ambermd.org/netcdf/nctraj.xhtml
        """
        # Set attributes.
        setattr(
            self._handle,
            "title",
            f"CREATED at {datetime.now()} on {socket.gethostname()}",
        )
        setattr(self._handle, "application", "Omnia")
        setattr(self._handle, "program", "MDTraj")
        setattr(self._handle, "programVersion", mdtraj.__version__)
        setattr(self._handle, "Conventions", "AMBER")
        setattr(self._handle, "ConventionVersion", "1.0")

        # set the dimensions
        # unlimited number of frames in trajectory
        self._handle.createDimension("frame", 0)
        # number of spatial coordinates
        self._handle.createDimension("spatial", 3)
        # number of atoms
        self._handle.createDimension("atom", n_atoms)

        if set_cell:
            # three spatial coordinates for the length of the unit cell
            self._handle.createDimension("cell_spatial", 3)
            # three spatial coordinates for the angles that define the shape
            # of the unit cell
            self._handle.createDimension("cell_angular", 3)
            # length of the longest string used for a label
            self._handle.createDimension("label", 5)

            # Define variables to store unit cell data
            cell_lengths = self._handle.createVariable(
                "cell_lengths",
                "d",
                ("frame", "cell_spatial"),
            )
            setattr(cell_lengths, "units", "angstrom")
            cell_angles = self._handle.createVariable(
                "cell_angles",
                "d",
                ("frame", "cell_angular"),
            )
            setattr(cell_angles, "units", "degree")

            self._handle.createVariable("cell_spatial", "c", ("cell_spatial",))
            self._handle.variables["cell_spatial"][0] = "a"
            self._handle.variables["cell_spatial"][1] = "b"
            self._handle.variables["cell_spatial"][2] = "c"

            self._handle.createVariable("cell_angular", "c", ("cell_spatial", "label"))
            self._handle.variables["cell_angular"][0] = "alpha"
            self._handle.variables["cell_angular"][1] = "beta "
            self._handle.variables["cell_angular"][2] = "gamma"

        if set_time:
            # Define coordinates and snapshot times.
            frame_times = self._handle.createVariable("time", "f", ("frame",))
            setattr(frame_times, "units", "picosecond")

        if set_coordinates:
            frame_coordinates = self._handle.createVariable(
                "coordinates",
                "f",
                ("frame", "atom", "spatial"),
            )
            setattr(frame_coordinates, "units", "angstrom")

            self._handle.createVariable("spatial", "c", ("spatial",))
            self._handle.variables["spatial"][0] = "x"
            self._handle.variables["spatial"][1] = "y"
            self._handle.variables["spatial"][2] = "z"

    def seek(self, offset, whence=0):
        """Move to a new file position

        Parameters
        ----------
        offset : int
            A number of frames.
        whence : {0, 1, 2}
            0: offset from start of file, offset should be >=0.
            1: move relative to the current position, positive or negative
            2: move relative to the end of file, offset should be <= 0.
            Seeking beyond the end of a file is not supported
        """
        if whence == 0 and offset >= 0:
            self._frame_index = offset
        elif whence == 1:
            self._frame_index = self._frame_index + offset
        elif whence == 2 and offset <= 0:
            self._frame_index = self.n_frames + offset
        else:
            raise OSError("Invalid argument")

    def tell(self):
        """Current file position

        Returns
        -------
        offset : int
            The current frame in the file.
        """
        return int(self._frame_index)

    def close(self):
        """Close the NetCDF file handle"""
        if not self._closed and hasattr(self, "_handle"):
            self._handle.close()

        self._closed = True

    def __enter__(self):
        # supports the context manager protocol
        return self

    def __exit__(self, *exc_info):
        # supports the context manager protocol
        self.close()

    def __del__(self):
        self.close()

    def __len__(self):
        if self._closed:
            raise ValueError("I/O operation on closed file")
        return self.n_frames



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2014 Stanford University and the Authors
#
# Authors: Jason Swails
# Contributors:
#
# This code for reading Amber restart and inpcrd files was taken from ParmEd,
# which is released under the GNU Lesser General Public License
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

"""
This module provides the ability to read Amber inpcrd/restart files as well as
Amber NetCDF restart files. This code was taken from ParmEd and simplified by
removing the functionality that is not needed.
"""

import os
import warnings
from math import ceil

import numpy as np
from packaging.version import Version

import mdtraj
from mdtraj.formats.registry import FormatRegistry
from mdtraj.utils import cast_indices, ensure_type, import_, in_units_of

__all__ = [
    "AmberRestartFile",
    "load_restrt",
    "AmberNetCDFRestartFile",
    "load_ncrestrt",
]


@FormatRegistry.register_loader(".rst7")
@FormatRegistry.register_loader(".restrt")
@FormatRegistry.register_loader(".inpcrd")
def load_restrt(filename, top=None, atom_indices=None):
    """Load an AMBER ASCII restart/inpcrd file. Since this file doesn't contain
    information to specify the topology, you need to supply a topology

    Parameters
    ----------
    filename : path-like
        name of the AMBER restart file
    top : {str, Trajectory, Topology}
        Pass in either the path to a file containing topology information (e.g.,
        a PDB, an AMBER prmtop, or certain types of Trajectory objects) to
        supply the necessary topology information that is not present in these
        files
    atom_indices : array_like, optional
        If not None, then read only a subset of the atoms coordinates from the
        file.

    Returns
    -------
    trajectory : md.Trajectory
        The resulting trajectory, as an md.Trajectory object

    See Also
    --------
    mdtraj.AmberRestartFile : Low level interface to AMBER restart files
    """
    from mdtraj.core.trajectory import _parse_topology

    topology = _parse_topology(top)
    atom_indices = cast_indices(atom_indices)

    with AmberRestartFile(filename) as f:
        return f.read_as_traj(topology, atom_indices=atom_indices)


@FormatRegistry.register_fileobject(".rst7")
@FormatRegistry.register_fileobject(".restrt")
@FormatRegistry.register_fileobject(".inpcrd")
class AmberRestartFile:
    """Interface for reading and writing AMBER ASCII restart files. This is a
    file-like object, that supports both reading and writing depending on the
    `mode` flag.  It implements the context manager protocol, so you can also
    use it with the python 'with' statement.

    Parameters
    ----------
    filename : path-like
        The name of the file to open
    mode : {'r', 'w'}, default='r'
        The mode in which to open the file. Valid options are 'r' or 'w' for
        'read' or 'write'
    force_overwrite : bool, default=False
        In write mode, if a file named `filename` already exists, clobber it and
        overwrite it

    See Also
    --------
    md.AmberNetCDFRestartFile : Low level interface to AMBER NetCDF-format restart files
    """

    distance_unit = "angstroms"

    def __init__(self, filename, mode="r", force_overwrite=True):
        self._closed = True
        self._mode = mode
        self._filename = filename

        if mode not in ("r", "w"):
            raise ValueError("mode must be one of ['r', 'w']")

        if mode == "w" and not force_overwrite and os.path.exists(filename):
            raise OSError('"%s" already exists' % filename)

        if mode == "w":
            self._needs_initialization = True
            self._handle = open(filename, mode)
            self._closed = False
        elif mode == "r":
            with open(filename, mode) as f:
                f.readline()
                words = f.readline().split()
                try:
                    self._n_atoms = int(words[0])
                except (IndexError, ValueError):
                    raise TypeError(
                        '"%s" is not a recognized Amber restart' % filename,
                    )
            self._needs_initialization = False
        else:
            raise RuntimeError()

    @property
    def n_atoms(self):
        self._validate_open()
        if self._needs_initialization:
            raise OSError("The file is uninitialized")
        return self._n_atoms

    @property
    def n_frames(self):
        return 1  # always 1 frame

    def _validate_open(self):
        if self._closed:
            raise OSError("The file is closed.")

    def _parse(self, lines):
        """Parses the file"""
        self._time = None
        try:
            words = lines[1].split()
            self._n_atoms = natom = int(words[0])
        except (IndexError, ValueError):
            raise TypeError("not a recognized Amber restart")

        time = None
        if len(words) >= 2:
            time = float(words[1])

        # variable _ throughout was hasvels, but velocies are not tracked
        lines_per_frame = int(ceil(natom / 2))
        if len(lines) == lines_per_frame + 2:
            hasbox = _ = False
        elif natom in (1, 2) and len(lines) == 4:
            # This is the _only_ case where line counting does not work -- there
            # is either 1 or 2 atoms and there are 4 lines. The 1st 3 lines are
            # the title, natom/time, and coordinates. The 4th are almost always
            # velocities since it's hard to have a periodic system this small.
            # However, velocities (which are scaled down by 20.445) have a ~0%
            # chance of being 60+, so we can pretty easily tell if the last line
            # has box dimensions and angles or velocities. I cannot envision a
            # plausible scenario where the detection here will ever fail
            line = lines[3]
            if natom == 1:
                tmp = [line[i : i + 12] for i in range(0, 72, 12) if line[i : i + 12].strip()]
                if len(tmp) == 3:
                    _ = True
                    hasbox = False
                elif len(tmp) == 6:
                    hasbox = True
                    _ = False
                else:
                    raise TypeError("not a recognized Amber restart")
            else:
                # Ambiguous case
                tmp = [float(line[i : i + 12]) >= 60.0 for i in range(0, 72, 12)]
                if any(tmp):
                    hasbox = True
                    _ = False
                else:
                    _ = True
                    hasbox = False
        elif len(lines) == lines_per_frame + 3:
            hasbox = True
            _ = False
        elif len(lines) == 2 * lines_per_frame + 2:
            hasbox = False
            _ = True
        elif len(lines) == 2 * lines_per_frame + 3:
            hasbox = _ = True
        else:
            raise TypeError(
                "Badly formatted restart file. Has %d lines for %d atoms" % (len(lines), natom),
            )

        coordinates = np.zeros((1, natom, 3))
        if time is None:
            time = np.zeros(1)
        else:
            time = np.asarray((time,))

        # Fill the coordinates
        for i in range(lines_per_frame):
            line = lines[i + 2]  # Skip first two lines
            i2 = i * 2
            coordinates[0, i2, :] = [float(line[j : j + 12]) for j in range(0, 36, 12)]
            i2 += 1
            if i2 < natom:
                coordinates[0, i2, :] = [float(line[j : j + 12]) for j in range(36, 72, 12)]
        if hasbox:
            cell_lengths = np.zeros((1, 3))
            cell_angles = np.zeros((1, 3))
            line = lines[-1]
            cell_lengths[0, :] = [float(line[i : i + 12]) for i in range(0, 36, 12)]
            cell_angles[0, :] = [float(line[i : i + 12]) for i in range(36, 72, 12)]
        else:
            cell_lengths = cell_angles = None

        return coordinates, time, cell_lengths, cell_angles

    def read_as_traj(self, topology, atom_indices=None):
        """Read an AMBER ASCII restart file as a trajectory.

        Parameters
        ----------
        topology : Topology
            The system topology
        atom_indices : array_like, optional
            If not none, then read only a subset of the atoms coordinates from the
            file. This may be slightly slower than the standard read because it required
            an extra copy, but will save memory.

        Returns
        -------
        trajectory : Trajectory
            A trajectory object with 1 frame created from the file.
        """
        from mdtraj.core.trajectory import Trajectory

        if atom_indices is not None:
            topology = topology.subset(atom_indices)

        xyz, time, cell_lengths, cell_angles = self.read(atom_indices=atom_indices)
        xyz = in_units_of(
            xyz,
            self.distance_unit,
            Trajectory._distance_unit,
            inplace=True,
        )
        cell_lengths = in_units_of(
            cell_lengths,
            self.distance_unit,
            Trajectory._distance_unit,
            inplace=True,
        )

        return Trajectory(
            xyz=xyz,
            topology=topology,
            time=time,
            unitcell_lengths=cell_lengths,
            unitcell_angles=cell_angles,
        )

    def read(self, atom_indices=None):
        """Read data from an AMBER ASCII restart file

        Parameters
        ----------
        atom_indices : np.ndarray, dtype=int, optional
            The specific indices of the atoms you'd like to retrieve. If not
            supplied, all of the atoms will be retrieved.

        Returns
        -------
        coordinates : np.ndarray, shape=(1, n_atoms, 3)
            The cartesian coordinates of the atoms, in units of angstroms. These
            files only ever contain 1 frame
        time : np.ndarray, None
            The time corresponding to the frame, in units of picoseconds, or
            None if no time information is present
        cell_lengths : np.ndarray, None
            The lengths (a, b, c) of the unit cell for the frame in angstroms,
            or None if the information is not present in the file
        cell_angles : np.ndarray, None
            The angles (\alpha, \beta, \\gamma) defining the unit cell for each
            frame, or None if the information is not present in the file.
        """
        if self._mode != "r":
            raise OSError(
                "The file was opened in mode=%s. Reading is not allowed." % self._mode,
            )

        with open(self._filename) as f:
            lines = f.readlines()

        coordinates, time, cell_lengths, cell_angles = self._parse(lines)

        if atom_indices is not None:
            atom_slice = ensure_type(
                atom_indices,
                dtype=int,
                ndim=1,
                name="atom_indices",
                warn_on_cast=False,
            )
            if not np.all(atom_slice) >= 0:
                raise ValueError("Entries in atom_slice must be >= 0")
            coordinates = coordinates[:, atom_slice, :]

        return coordinates, time, cell_lengths, cell_angles

    def write(
        self,
        coordinates,
        time=None,
        cell_lengths=None,
        cell_angles=None,
    ):
        """Write one frame of a MD trajectory to disk in the AMBER ASCII restart
        file format.

        Parameters
        ----------
        coordinates : np.ndarray, dtype=np.float32, shape=([1,] n_atoms, 3)
            The cartesian coordinates of each atom, in units of angstroms. Must
            be only a single frame (shape can be (1,N,3) or (N,3) where N is
            the number of atoms)
        time : array-like with 1 element or float, optional
            The time corresponding to this frame. If not specified, a place
            holder of 0 will be written
        cell_lengths : np.ndarray, dtype=np.double, shape=([1,] 3)
            The lengths (a,b,c) of the unit cell for the frame in Angstroms
        cell_angles : np.ndarray, dtype=np.double, shape=([1,] 3)
            The angles between the unit cell vectors for the frame in Degrees
        """
        if self._mode != "w":
            raise OSError(
                "The file was opened in mode=%s. Writing not allowed." % self._mode,
            )
        if not self._needs_initialization:
            # Must have already been written -- can only write once
            raise RuntimeError(
                "restart file has already been written -- can only write one frame to restart files.",
            )
        # These are no-ops.
        # coordinates = in_units_of(coordinates, None, 'angstroms')
        # time = in_units_of(time, None, 'picoseconds')
        # cell_lengths = in_units_of(cell_lengths, None, 'angstroms')
        # cell_angles = in_units_of(cell_angles, None, 'degrees')

        # typecheck all of the input arguments rigorously
        coordinates = ensure_type(
            coordinates,
            np.float32,
            3,
            "coordinates",
            length=None,
            can_be_none=False,
            shape=(1, None, 3),
            warn_on_cast=False,
            add_newaxis_on_deficient_ndim=True,
        )
        n_frames, self._n_atoms = coordinates.shape[0], coordinates.shape[1]
        if n_frames != 1:
            raise ValueError("Can only write 1 frame to a restart file!")
        if time is not None:
            try:
                if isinstance(time, np.ndarray):
                    time = float(time.item())
                else:
                    time = float(time)
            except TypeError:
                raise TypeError("Can only provide a single time")
        else:
            time = 0.0
        cell_lengths = ensure_type(
            cell_lengths,
            np.float64,
            2,
            "cell_lengths",
            length=1,
            can_be_none=True,
            warn_on_cast=False,
            add_newaxis_on_deficient_ndim=True,
        )
        cell_angles = ensure_type(
            cell_angles,
            np.float64,
            2,
            "cell_angles",
            length=1,
            can_be_none=True,
            warn_on_cast=False,
            add_newaxis_on_deficient_ndim=True,
        )
        if (cell_lengths is None and cell_angles is not None) or (cell_lengths is not None and cell_angles is None):
            prov, negl = "cell_lengths", "cell_angles"
            if cell_lengths is None:
                prov, negl = negl, prov
            raise ValueError(
                f'You provided the variable "{prov}" but did not '
                f'provide "{negl}". Either provide both or neither -- '
                "one without the other is meaningless.",
            )

        self._handle.write(
            "Amber restart file (without velocities) written by MDTraj\n",
        )
        self._handle.write("%5d%15.7e\n" % (self._n_atoms, time))
        fmt = "%12.7f%12.7f%12.7f"
        for i in range(self._n_atoms):
            acor = coordinates[0, i, :]
            self._handle.write(fmt % (acor[0], acor[1], acor[2]))
            if i % 2 == 1:
                self._handle.write("\n")
        if self._n_atoms % 2 == 1:
            self._handle.write("\n")
        if cell_lengths is not None:
            self._handle.write(
                fmt
                % (
                    cell_lengths[0, 0],
                    cell_lengths[0, 1],
                    cell_lengths[0, 2],
                ),
            )
            self._handle.write(
                fmt
                % (
                    cell_angles[0, 0],
                    cell_angles[0, 1],
                    cell_angles[0, 2],
                )
                + "\n",
            )
        self._handle.flush()

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()

    def close(self):
        if not self._closed and hasattr(self, "_handle"):
            self._handle.close()
        self._closed = True

    def __del__(self):
        self.close()

    def __len__(self):
        return 1  # All restarts have only 1 frame


@FormatRegistry.register_loader(".ncrst")
def load_ncrestrt(filename, top=None, atom_indices=None):
    """Load an AMBER NetCDF restart/inpcrd file. Since this file doesn't
    contain information to specify the topology, you need to supply a topology

    Parameters
    ----------
    filename : path-like
        name of the AMBER restart file
    top : {path-like, Trajectory, Topology}
        Pass in either the path to a file containing topology information (e.g.,
        a PDB, an AMBER prmtop, or certain types of Trajectory objects) to
        supply the necessary topology information that is not present in these
        files
    atom_indices : array_like, optional
        If not None, then read only a subset of the atoms coordinates from the
        file.

    Returns
    -------
    trajectory : md.Trajectory
        The resulting trajectory, as an md.Trajectory object

    See Also
    --------
    mdtraj.AmberRestartFile : Low level interface to AMBER restart files
    """
    from mdtraj.core.trajectory import _parse_topology

    topology = _parse_topology(top)
    atom_indices = cast_indices(atom_indices)

    with AmberNetCDFRestartFile(filename) as f:
        return f.read_as_traj(topology, atom_indices=atom_indices)


@FormatRegistry.register_fileobject(".ncrst")
class AmberNetCDFRestartFile:
    """Interface for reading and writing AMBER NetCDF files. This is a file-like
    object, that supports both reading and writing depending on the `mode` flag.
    It implements the context manager protocol, so you can also use it with the
    python 'with' statement.

    Parameters
    ----------
    filename : path-like
        The name of the file to open
    mode : {'r', 'w'}, default='r'
        The mode in which to open the file. Valid options are 'r' or 'w' for
        'read' or 'write'
    force_overwrite : bool, default=False
        In write mode, if a file named `filename` already exists, clobber it and
        overwrite it
    """

    distance_unit = "angstroms"

    def __init__(self, filename, mode="r", force_overwrite=False):
        self._closed = True
        self._mode = mode
        if Version(import_("scipy.version").short_version) < Version("0.12.0"):
            raise ImportError(
                "MDTraj NetCDF support requires scipy>=0.12.0. You have %s" % import_("scipy.version").short_version,
            )
        netcdf = import_("scipy.io").netcdf_file

        if mode not in ("r", "w"):
            raise ValueError("mode must be one of ['r', 'w']")

        if mode == "w" and not force_overwrite and os.path.exists(filename):
            raise OSError('"%s" already exists' % filename)

        # AMBER uses the NetCDF3 format, with 64 bit encodings, which for
        # scipy.io.netcdf_file is "version=2"
        self._handle = netcdf(filename, mode=mode, version=2)
        self._closed = False
        if mode == "w":
            self._needs_initialization = True
        elif mode == "r":
            self._needs_initialization = False
        else:
            raise RuntimeError()

    @property
    def n_atoms(self):
        self._validate_open()
        if self._needs_initialization:
            raise OSError("The file is uninitialized")
        return self._handle.dimensions["atom"]

    @property
    def n_frames(self):
        return 1  # always 1 frame

    def _validate_open(self):
        if self._closed:
            raise OSError("The file is closed.")

    def read_as_traj(self, topology, atom_indices=None):
        """Read an AMBER ASCII restart file as a trajectory.

        Parameters
        ----------
        topology : Topology
            The system topology
        atom_indices : array_like, optional
            If not none, then read only a subset of the atoms coordinates from the
            file. This may be slightly slower than the standard read because it required
            an extra copy, but will save memory.

        Returns
        -------
        trajectory : Trajectory
            A trajectory object with 1 frame created from the file.
        """
        from mdtraj.core.trajectory import Trajectory

        if atom_indices is not None:
            topology = topology.subset(atom_indices)

        xyz, time, cell_lengths, cell_angles = self.read(atom_indices=atom_indices)
        xyz = in_units_of(
            xyz,
            self.distance_unit,
            Trajectory._distance_unit,
            inplace=True,
        )
        cell_lengths = in_units_of(
            cell_lengths,
            self.distance_unit,
            Trajectory._distance_unit,
            inplace=True,
        )

        return Trajectory(
            xyz=xyz,
            topology=topology,
            time=time,
            unitcell_lengths=cell_lengths,
            unitcell_angles=cell_angles,
        )

    def read(self, atom_indices=None):
        """Read data from an AMBER NetCDF restart file

        Parameters
        ----------
        atom_indices : np.ndarray, dtype=int, optional
            The specific indices of the atoms you'd like to retrieve. If not
            supplied, all of the atoms will be retrieved.

        Returns
        -------
        coordinates : np.ndarray, shape=(1, n_atoms, 3)
            The cartesian coordinates of the atoms, in units of angstroms. These
            files only ever contain 1 frame
        time : np.ndarray, None
            The time corresponding to the frame, in units of picoseconds, or
            None if no time information is present
        cell_lengths : np.ndarray, None
            The lengths (a, b, c) of the unit cell for the frame in angstroms,
            or None if the information is not present in the file
        cell_angles : np.ndarray, None
            The angles (\alpha, \beta, \\gamma) defining the unit cell for each
            frame, or None if the information is not present in the file.

        Notes
        -----
        If the file is not a NetCDF file with the appropriate convention, a
        TypeError is raised. If variables that are needed do not exist or if
        illegal values are passed in for parameters, ValueError is raised. If
        I/O errors occur, IOError is raised.
        """
        if self._mode != "r":
            raise OSError(
                "The file was opened in mode=%s. Reading is not allowed." % self._mode,
            )

        if self._closed:
            raise OSError("The file has been closed!")

        if "coordinates" not in self._handle.variables:
            raise ValueError("No coordinates found in the NetCDF file.")

        # Check that conventions are correct
        try:
            conventions = self._handle.Conventions.decode("ascii")
        except UnicodeDecodeError:
            raise TypeError("NetCDF file does not have correct Conventions")
        try:
            convention_version = self._handle.ConventionVersion.decode("ascii")
        except UnicodeDecodeError:
            raise ValueError("NetCDF file does not have correct ConventionVersion")
        except AttributeError:
            raise TypeError("NetCDF file does not have ConventionVersion")
        if not hasattr(self._handle, "Conventions") or conventions != "AMBERRESTART":
            raise TypeError("NetCDF file does not have correct Conventions")
        if convention_version != "1.0":
            raise ValueError(
                "NetCDF restart has ConventionVersion %s. Only Version 1.0 is supported." % convention_version,
            )
        if atom_indices is not None:
            atom_slice = ensure_type(
                atom_indices,
                dtype=int,
                ndim=1,
                name="atom_indices",
                warn_on_cast=False,
            )
            if not np.all(atom_slice) >= 0:
                raise ValueError("Entries in atom_slice must be >= 0")
            coordinates = self._handle.variables["coordinates"][atom_slice, :]
        else:
            coordinates = self._handle.variables["coordinates"][:, :]

        # Get unit cell parameters
        if "cell_lengths" in self._handle.variables:
            cell_lengths = self._handle.variables["cell_lengths"][:]
        else:
            cell_lengths = None
        if "cell_angles" in self._handle.variables:
            cell_angles = self._handle.variables["cell_angles"][:]
        else:
            cell_angles = None

        if cell_lengths is None and cell_angles is not None:
            warnings.warn("cell_lengths were found, but no cell_angles")
        if cell_lengths is not None and cell_angles is None:
            warnings.warn("cell_angles were found, but no cell_lengths")

        if "time" in self._handle.variables:
            time = self._handle.variables["time"].getValue()
        else:
            time = None

        # scipy.io.netcdf variables are mem-mapped, and are only backed by valid
        # memory while the file handle is open. This is _bad_ because we need to
        # support the user opening the file, reading the coordinates, and then
        # closing it, and still having the coordinates be a valid memory
        # segment.
        # https://github.com/mdtraj/mdtraj/issues/440
        if coordinates is not None and not coordinates.flags["WRITEABLE"]:
            coordinates = np.array(coordinates, copy=True)
        if cell_lengths is not None and not cell_lengths.flags["WRITEABLE"]:
            cell_lengths = np.array(cell_lengths, copy=True)
        if cell_angles is not None and not cell_angles.flags["WRITEABLE"]:
            cell_angles = np.array(cell_angles, copy=True)

        # The leading frame dimension is missing on all of these arrays since
        # restart files have only one frame. Reshape them to add this extra
        # dimension
        coordinates = coordinates[np.newaxis, :]
        if cell_lengths is not None:
            cell_lengths = cell_lengths[np.newaxis, :]
        if cell_angles is not None:
            cell_angles = cell_angles[np.newaxis, :]
        if time is not None:
            time = np.asarray([time])

        return coordinates, time, cell_lengths, cell_angles

    def write(
        self,
        coordinates,
        time=None,
        cell_lengths=None,
        cell_angles=None,
    ):
        """Write one frame of a MD trajectory to disk in the AMBER NetCDF
        restart file format.

        Parameters
        ----------
        coordinates : np.ndarray, dtype=np.float32, shape=([1,] n_atoms, 3)
            The cartesian coordinates of each atom, in units of angstroms. Must
            be only a single frame (shape can be (1,N,3) or (N,3) where N is
            the number of atoms)
        time : array-like with 1 element or float, optional
            The time corresponding to this frame. If not specified, a place
            holder of 0 will be written
        cell_lengths : np.ndarray, dtype=np.double, shape=([1,] 3)
            The lengths (a,b,c) of the unit cell for the frame in Angstroms
        cell_angles : np.ndarray, dtype=np.double, shape=([1,] 3)
            The angles between the unit cell vectors for the frame in Degrees

        Notes
        -----
        You must only have one frame to write to this file.
        """
        if self._mode != "w":
            raise OSError(
                "The file was opened in mode=%s. Writing not allowed." % self._mode,
            )
        if not self._needs_initialization:
            # Must have already been written -- can only write once
            raise RuntimeError(
                "NetCDF restart file has already been written -- can only write one frame to restart files.",
            )
        # these are no-ops
        # coordinates = in_units_of(coordinates, None, 'angstroms')
        # time = in_units_of(time, None, 'picoseconds')
        # cell_lengths = in_units_of(cell_lengths, None, 'angstroms')
        # cell_angles = in_units_of(cell_angles, None, 'degrees')

        # typecheck all of the input arguments rigorously
        coordinates = ensure_type(
            coordinates,
            np.float32,
            3,
            "coordinates",
            length=None,
            can_be_none=False,
            shape=(1, None, 3),
            warn_on_cast=False,
            add_newaxis_on_deficient_ndim=True,
        )
        n_frames, n_atoms = coordinates.shape[0], coordinates.shape[1]
        if n_frames != 1:
            raise ValueError("Can only write 1 frame to a restart file!")
        if time is not None:
            try:
                if isinstance(time, np.ndarray):
                    time = float(time.item())
                else:
                    time = float(time)
            except TypeError:
                raise TypeError("Can only provide a single time")
        else:
            time = 0.0
        cell_lengths = ensure_type(
            cell_lengths,
            np.float64,
            2,
            "cell_lengths",
            length=1,
            can_be_none=True,
            warn_on_cast=False,
            add_newaxis_on_deficient_ndim=True,
        )
        cell_angles = ensure_type(
            cell_angles,
            np.float64,
            2,
            "cell_angles",
            length=1,
            can_be_none=True,
            warn_on_cast=False,
            add_newaxis_on_deficient_ndim=True,
        )
        if (cell_lengths is None and cell_angles is not None) or (cell_lengths is not None and cell_angles is None):
            prov, negl = "cell_lengths", "cell_angles"
            if cell_lengths is None:
                prov, negl = negl, prov
            raise ValueError(
                f'You provided the variable "{prov}" but did not '
                f'provide "{negl}". Either provide both or neither -- '
                "one without the other is meaningless.",
            )

        self._initialize_headers(
            n_atoms=n_atoms,
            set_coordinates=True,
            set_time=(time is not None),
            set_cell=(cell_lengths is not None),
        )
        self._needs_initialization = False

        # Write the time, coordinates, and box info
        if time is not None:
            if isinstance(time, np.ndarray):
                self._handle.variables["time"][0] = float(time.item())
            else:
                self._handle.variables["time"][0] = float(time)
        self._handle.variables["coordinates"][:, :] = coordinates[0, :, :]
        if cell_lengths is not None:
            self._handle.variables["cell_angles"][:] = cell_angles[0, :]
            self._handle.variables["cell_lengths"][:] = cell_lengths[0, :]
        self.flush()

    def _initialize_headers(self, n_atoms, set_coordinates, set_time, set_cell):
        """Initialize the headers and convention properties of the NetCDF
        restart file
        """
        ncfile = self._handle
        ncfile.Conventions = "AMBERRESTART"
        ncfile.ConventionVersion = "1.0"
        ncfile.title = "NetCDF Restart file written by MDTraj w/out velocities"
        ncfile.application = "Omnia"
        ncfile.program = "MDTraj"
        ncfile.programVersion = mdtraj.__version__
        # Dimensions
        ncfile.createDimension("spatial", 3)
        ncfile.createDimension("atom", n_atoms)
        if set_cell:
            ncfile.createDimension("cell_spatial", 3)
            ncfile.createDimension("label", 5)
            ncfile.createDimension("cell_angular", 3)
        if set_time:
            ncfile.createDimension("time", 1)
        # Variables
        v = ncfile.createVariable("spatial", "c", ("spatial",))
        v[:] = np.asarray(list("xyz"))
        v = ncfile.createVariable("coordinates", "d", ("atom", "spatial"))
        v.units = "angstrom"
        if set_cell:
            v = ncfile.createVariable(
                "cell_angular",
                "c",
                ("cell_angular", "label"),
            )
            v[0] = np.asarray(list("alpha"))
            v[1] = np.asarray(list("beta "))
            v[2] = np.asarray(list("gamma"))
            v = ncfile.createVariable("cell_spatial", "c", ("cell_spatial",))
            v[:] = np.asarray(list("abc"))
            v = ncfile.createVariable("cell_lengths", "d", ("cell_spatial",))
            v.units = "angstrom"
            v = ncfile.createVariable("cell_angles", "d", ("cell_angular",))
            v.units = "degree"
        if set_time:
            v = ncfile.createVariable("time", "d", ("time",))
            v.units = "picosecond"
        self.flush()

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()

    def close(self):
        if not self._closed and hasattr(self, "_handle"):
            self._handle.close()
        self._closed = True

    def __del__(self):
        self.close()

    def __len__(self):
        return 1  # All restarts have only 1 frame

    def flush(self):
        self._validate_open()
        if self._mode != "w":
            raise OSError("Cannot flush a file opened for reading")
        self._handle.flush()



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors: Kyle A. Beauchamp
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


from .angle import *
from .contact import *
from .dihedral import *
from .distance import *
from .drid import *
from .dssp import *
from .hbond import *
from .neighborlist import *
from .neighbors import *
from .order import *
from .pi_stacking import *
from .rdf import *
from .rg import *
from .sasa import *
from .shape import *
from .thermodynamic_properties import *

__all__ = [
    "baker_hubbard",
    "shrake_rupley",
    "kabsch_sander",
    "compute_distances",
    "compute_distances_t",
    "compute_displacements",
    "compute_angles",
    "compute_dihedrals",
    "compute_phi",
    "compute_psi",
    "compute_chi1",
    "compute_chi2",
    "compute_chi3",
    "compute_chi4",
    "compute_chi5",
    "compute_omega",
    "compute_rg",
    "compute_contacts",
    "compute_drid",
    "compute_center_of_mass",
    "compute_center_of_geometry",
    "wernet_nilsson",
    "pi_stacking",
    "compute_dssp",
    "compute_neighbors",
    "compute_neighborlist",
    "compute_rdf",
    "compute_rdf_t",
    "compute_nematic_order",
    "compute_inertia_tensor",
    "compute_gyration_tensor",
    "find_closest_contact",
    "compute_directors",
    "principal_moments",
    "asphericity",
    "acylindricity",
    "relative_shape_antisotropy",
    "dipole_moments",
    "static_dielectric",
    "isothermal_compressability_kappa_T",
    "thermal_expansion_alpha_P",
    "density",
]



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

import os
import shutil
import subprocess
import sys
import tempfile

import numpy as np
import pytest
from _pytest.monkeypatch import MonkeyPatch

import mdtraj as md
from mdtraj.formats import NetCDFTrajectoryFile
from mdtraj.testing import eq

needs_cpptraj = pytest.mark.skipif(
    shutil.which("cpptraj") is None,
    reason="This test requires cpptraj from AmberTools to be installed (http://ambermd.org)",
)

fd, temp = tempfile.mkstemp(suffix=".nc")
fd2, temp2 = tempfile.mkstemp(suffix=".nc")


class TestNetCDFNetCDF4:
    """
    This class contains all the tests that we would also want to run with scipy.
    Now a class so we can subclass it for later.
    """

    def teardown_module(self, module):
        """remove the temporary file created by tests in this file
        this gets automatically called by pytest"""
        os.close(fd)
        os.close(fd2)
        os.unlink(temp)
        os.unlink(temp2)

    def test_read_after_close(self, get_fn):
        """Default test using netCDF4"""
        f = NetCDFTrajectoryFile(get_fn("mdcrd.nc"))
        assert eq(f.n_atoms, 223)
        assert eq(f.n_frames, 101)

        f.close()

        # should be an IOError if you read a file that's closed
        with pytest.raises(IOError):
            f.read()

    def test_shape(self, get_fn):
        """Default test using netCDF4"""
        with NetCDFTrajectoryFile(get_fn("mdcrd.nc")) as netcdf_file:
            xyz, time, boxlength, boxangles = netcdf_file.read()

        assert eq(xyz.shape, (101, 223, 3))
        assert eq(time.shape, (101,))
        assert eq(boxlength, None)
        assert eq(boxangles, None)

    def test_read_chunk_1(self, get_fn):
        """Default test using netCDF4"""
        with NetCDFTrajectoryFile(get_fn("mdcrd.nc")) as file:
            a, b, c, d = file.read(10)
            e, f, g, h = file.read()

            assert eq(len(a), 10)
            assert eq(len(b), 10)

            assert eq(len(e), 101 - 10)
            assert eq(len(f), 101 - 10)

        with NetCDFTrajectoryFile(get_fn("mdcrd.nc")) as file:
            xyz = file.read()[0]

        assert eq(a, xyz[0:10])
        assert eq(e, xyz[10:])

    def test_read_chunk_2(self, get_fn):
        """Default test using netCDF4"""

        with NetCDFTrajectoryFile(get_fn("mdcrd.nc")) as file:
            a, b, c, d = file.read(10)
            e, f, g, h = file.read(100000000000)

            assert eq(len(a), 10)
            assert eq(len(b), 10)

            assert eq(len(e), 101 - 10)
            assert eq(len(f), 101 - 10)

        with NetCDFTrajectoryFile(get_fn("mdcrd.nc")) as file:
            xyz = file.read()[0]

        assert eq(a, xyz[0:10])
        assert eq(e, xyz[10:])

    def test_read_chunk_3(self, get_fn):
        """Default test using netCDF4"""
        # too big of a chunk should not be an issue
        with NetCDFTrajectoryFile(get_fn("mdcrd.nc")) as file:
            a = file.read(1000000000)
        with NetCDFTrajectoryFile(get_fn("mdcrd.nc")) as file:
            b = file.read()

        eq(a[0], b[0])

    def test_read_write_1(self):
        """Default test using netCDF4"""
        xyz = np.random.randn(100, 3, 3)
        time = np.random.randn(100)
        boxlengths = np.random.randn(100, 3)
        boxangles = np.random.randn(100, 3)

        with NetCDFTrajectoryFile(temp, "w", force_overwrite=True) as f:
            f.write(xyz, time, boxlengths, boxangles)

        with NetCDFTrajectoryFile(temp) as f:
            a, b, c, d = f.read()
            assert eq(a, xyz)
            assert eq(b, time)
            assert eq(c, boxlengths)
            assert eq(d, boxangles)

    def test_read_write_2(self, get_fn):
        """Default test using netCDF4"""
        xyz = np.random.randn(5, 22, 3)
        time = np.random.randn(5)

        with NetCDFTrajectoryFile(temp, "w", force_overwrite=True) as f:
            f.write(xyz, time)

        with NetCDFTrajectoryFile(temp) as f:
            rcoord, rtime, rlengths, rangles = f.read()
            assert eq(rcoord, xyz)
            assert eq(rtime, time)
            assert eq(rlengths, None)
            assert eq(rangles, None)

        t = md.load(temp, top=get_fn("native.pdb"))
        eq(t.unitcell_angles, None)
        eq(t.unitcell_lengths, None)

    def test_ragged_1(self):
        """Default test using netCDF4"""
        # try first writing no cell angles/lengths, and then adding some
        xyz = np.random.randn(100, 3, 3)
        time = np.random.randn(100)
        cell_lengths = np.random.randn(100, 3)
        cell_angles = np.random.randn(100, 3)

        with NetCDFTrajectoryFile(temp, "w", force_overwrite=True) as f:
            f.write(xyz, time)
            with pytest.raises(ValueError):
                f.write(xyz, time, cell_lengths, cell_angles)

    def test_ragged_2(self):
        """Default test using netCDF4"""
        # try first writing no cell angles/lengths, and then adding some
        xyz = np.random.randn(100, 3, 3)
        time = np.random.randn(100)
        cell_lengths = np.random.randn(100, 3)
        cell_angles = np.random.randn(100, 3)

        # from mdtraj.formats import HDF5TrajectoryFile
        with NetCDFTrajectoryFile(temp, "w", force_overwrite=True) as f:
            f.write(xyz, time, cell_lengths, cell_angles)
            with pytest.raises(ValueError):
                f.write(xyz, time)

    def test_read_write_25(self):
        """Default test using netCDF4"""
        xyz = np.random.randn(100, 3, 3)
        time = np.random.randn(100)

        with NetCDFTrajectoryFile(temp, "w", force_overwrite=True) as f:
            f.write(xyz, time)
            f.write(xyz, time)

        with NetCDFTrajectoryFile(temp) as f:
            a, b, c, d = f.read()
            assert eq(a[0:100], xyz)
            assert eq(b[0:100], time)
            assert eq(c, None)
            assert eq(d, None)

            assert eq(a[100:], xyz)
            assert eq(b[100:], time)
            assert eq(c, None)
            assert eq(d, None)

    def test_write_3(self):
        """Default test using netCDF4"""
        with NetCDFTrajectoryFile(temp, "w", force_overwrite=True) as f:
            # you can't supply cell_lengths without cell_angles
            with pytest.raises(ValueError):
                f.write(np.random.randn(100, 3, 3), cell_lengths=np.random.randn(100, 3))
            # or the other way around
            with pytest.raises(ValueError):
                f.write(np.random.randn(100, 3, 3), cell_angles=np.random.randn(100, 3))

    def test_n_atoms(self):
        """Default test using netCDF4"""
        with NetCDFTrajectoryFile(temp, "w", force_overwrite=True) as f:
            f.write(np.random.randn(1, 11, 3))
        with NetCDFTrajectoryFile(temp) as f:
            eq(f.n_atoms, 11)

    def test_do_overwrite(self):
        """Default test using netCDF4"""
        with open(temp, "w") as f:
            f.write("a")

        with NetCDFTrajectoryFile(temp, "w", force_overwrite=True) as f:
            f.write(np.random.randn(10, 5, 3))

    def test_do_not_overwrite(self):
        """Default test using netCDF4"""
        with open(temp, "w") as f:
            f.write("a")

        with pytest.raises(IOError):
            with NetCDFTrajectoryFile(temp, "w", force_overwrite=False) as f:
                f.write(np.random.randn(10, 5, 3))

    def test_trajectory_save_load(self, get_fn):
        """Default test using netCDF4"""
        t = md.load(get_fn("native.pdb"))
        t.unitcell_lengths = 1 * np.ones((1, 3))
        t.unitcell_angles = 90 * np.ones((1, 3))

        t.save(temp)
        t2 = md.load(temp, top=t.topology)

        eq(t.xyz, t2.xyz)
        eq(t.unitcell_lengths, t2.unitcell_lengths)


class TestNetCDFScipy(TestNetCDFNetCDF4):
    """This inherits the TestNetCDFNetCDF4 class and run all tests with SciPy"""

    def setup_method(self, method):
        """Patching out netCDF4. This is the way to do it inside a class"""
        monkeypatch = MonkeyPatch()
        monkeypatch.setitem(sys.modules, "netCDF4", None)

    def teardown_method(self, method):
        """Undoing most changes, just in case."""
        monkeypatch = MonkeyPatch()
        monkeypatch.delitem(sys.modules, "netCDF4", None)


@needs_cpptraj
def test_cpptraj(get_fn):
    trj0 = md.load(get_fn("frame0.dcd"), top=get_fn("frame0.pdb"))
    trj0.save(temp)

    top = get_fn("frame0.pdb")
    subprocess.check_call(
        [
            "cpptraj",
            "-p",
            top,
            "-y",
            temp,
            "-x",
            temp2,
        ],
    )

    trj1 = md.load(temp, top=top)
    trj2 = md.load(temp2, top=top)

    np.testing.assert_array_almost_equal(trj0.xyz, trj2.xyz)
    np.testing.assert_array_almost_equal(trj1.xyz, trj2.xyz)
    np.testing.assert_array_almost_equal(
        trj0.unitcell_vectors,
        trj2.unitcell_vectors,
    )
    np.testing.assert_array_almost_equal(
        trj1.unitcell_vectors,
        trj2.unitcell_vectors,
    )

    np.testing.assert_array_almost_equal(trj0.time, trj1.time)
    np.testing.assert_array_almost_equal(trj0.time, trj2.time)
    np.testing.assert_array_almost_equal(trj1.time, trj2.time)
